package org.convey;

import javax.swing.*;

import org.jdom.*;
import org.jdom.input.*;

import org.jdom.output.*;
import java.io.*;
import java.util.*;

import java.awt.*;

/**
 * @author Andy Ames, Ed Braunhut
 * @version 1.0
 *
 * @param  modified  set to true at end of GraphicPanel's handleGraphicEventElement
 *                   method and set to false initially, and in GraphicPanel's openSVG
 *                   and saveSVG methods.
 */
public class CanvasPanel extends JPanel {

	protected GraphicPanel graphicPanel;

	protected ConveyFrame conveyFrame;

	protected boolean saved = false;

	protected boolean modified = false;

	protected boolean isNetworked = false;

	protected String savedFilename = "Untitled";

	public CanvasPanel(ConveyFrame conveyFrame) {
		super.setLayout(new BorderLayout(5, 5));

		this.graphicPanel = new GraphicPanel(this);

		this.conveyFrame = conveyFrame;

		initCenterComponent();
	}

	public void initCenterComponent() {
		super.add(this.graphicPanel, BorderLayout.CENTER);
	}

	public GraphicPanel getGraphicPanel() {
		return this.graphicPanel;
	}

	public ConveyFrame getConveyFrame() {
		return this.conveyFrame;
	}

	public void setModified() {
	 this.modified = true;
	}

	public void setNetworked() {
		this.isNetworked = true;
	}

	public void saveSVG(String filename) {
		Element root = new Element("svg");
		//using SVG 1.0 Specification currently (6.20.2003) supported
		DocType type = new DocType("svg", "-//W3C//DTD SVG 1.0//EN",
															 "http://www.w3.org/TR/2001/REC-SVG-20010904/DTD/svg10.dtd");
		Document doc = new Document(root, type);

		Iterator iter = conveyFrame.getSelectedGraphicPanel().graphicIterator();
		while (iter.hasNext()) {
			Graphic graphic = (Graphic) iter.next();
			Element graphicElement = graphic.toElement();
			root.addContent(graphicElement);
		}

		//FileOutputStream fos = null;
		try {
			//fos = new FileOutputStream("");
			XMLOutputter serializer = new XMLOutputter();
			serializer.setIndent("  ");
			serializer.setNewlines(true);
			//serializer.output(doc, fos);

			System.out.println(serializer.outputString(doc)); //<------------testing

			ObjectOutputStream os = new ObjectOutputStream(
						new FileOutputStream(filename));
			os.writeObject(serializer.outputString(doc));
			os.close();

		} catch (Exception e) {
			e.printStackTrace();
		}
		saved = true;
		modified = false;
		this.conveyFrame.updateButtons();
	}

	public void saveSVG(File file){
		if(file == null || file.getName().equals("")) {
			JOptionPane.showMessageDialog(this, "Invalid File Name", "Invalid File Name", JOptionPane.ERROR_MESSAGE);
		} else {
			//add .svg extension if user didn't specify one
			boolean noDot = true;
			for (int i = 0; i < file.getName().length(); i++) {
				if (file.getName().charAt(i) == '.') {
					noDot = false;
					break;
				}
			}
			if(noDot)
				file = new File(file.getPath() + ".svg");

			if(!file.exists() || JOptionPane.showConfirmDialog(this, "Overwrite existing file: " + file.getName() + "?",
					"Overwrite?", JOptionPane.YES_NO_CANCEL_OPTION,
					JOptionPane.QUESTION_MESSAGE) == JOptionPane.YES_OPTION) {

				Element root = new Element("svg");
				//using SVG 1.0 Specification currently (6.20.2003) supported
				DocType type = new DocType("svg", "-//W3C//DTD SVG 1.0//EN",
																	 "http://www.w3.org/TR/2001/REC-SVG-20010904/DTD/svg10.dtd");
				Document doc = new Document(root, type);

				Iterator iter = conveyFrame.getSelectedGraphicPanel().graphicIterator();
				while (iter.hasNext()) {
					Graphic graphic = (Graphic) iter.next();
					Element graphicElement = graphic.toElement();
					root.addContent(graphicElement);
				}

				FileOutputStream fos = null;
				try {
					fos = new FileOutputStream(file);
					XMLOutputter serializer = new XMLOutputter();
					serializer.setIndent("  ");
					serializer.setNewlines(true);
					serializer.output(doc, fos);
					System.out.println(serializer.outputString(doc));
				}
				catch (Exception e) {
					e.printStackTrace();
				}
				finally {
					if (fos != null) {
						try {
							fos.close();
						}
						catch (Exception e) {
							e.printStackTrace();
						}
					}
				}
			}
			saved = true;
			this.savedFilename = file.getPath();
			modified = false;

			this.conveyFrame.setTitle();
			this.conveyFrame.updateButtons();
		}
	}

	public void openSVG(File file) {
		DOMBuilder builder = new DOMBuilder(false);
		Document doc;
		try {
			doc = builder.build(file);
		} catch(Exception e) {
			e.printStackTrace();
			return;
		}
		Element root = doc.getRootElement();
		if(root == null)
			return;
		if(root.getName().equals("svg")) {
			this.graphicPanel.editClear();
			Iterator iter = root.getChildren().iterator();
			while(iter.hasNext()) {
				Element element = (Element)iter.next();
				Graphic g = this.graphicPanel.createGraphicFromSVGElement(element);
				if(g != null) {
					this.graphicPanel.addGraphic(g);
					this.graphicPanel.pushGraphicEvent(new NewGraphicEvent(g));
				}
				GraphicEvent ge = this.graphicPanel.popAllGraphicEvents();
				this.graphicPanel.addGraphicEvent(ge);
			}
		}
		saved = true;
		this.savedFilename = file.getPath();
		modified = false;

		this.conveyFrame.setTitle();
		this.conveyFrame.updateButtons();
	}

	public String getSavedFilename() {
		return this.savedFilename;
	}

}