package org.convey;

import java.awt.*;

import java.awt.event.*;

import javax.swing.*;

import java.util.*;

import java.io.*;

/**
 * @author Andy Ames
 * @version 0.2
 */
public class ChatDialog extends JDialog {

	protected transient int result;

	protected transient String server;

	protected DefaultComboBoxModel serverComboBoxModel;

	protected transient JComboBox serverComboBox;

	protected transient String user;

	protected Hashtable userComboBoxModels;

	protected transient JComboBox userComboBox;

	protected transient JTextField subjectTextField;

	public ChatDialog() {
		super(ConveyFrame.getConveyFrame(), "Chat", true);
	}

	public ChatDialog(Frame frame) {
		super(frame, "Chat", true);

		Object[] serverObjects = new Object[]{"jabber.at", "jabber.com", "jabber.org"};
		this.serverComboBoxModel = new DefaultComboBoxModel(serverObjects);
		int i = this.serverComboBoxModel.getSize();

		this.userComboBoxModels = new Hashtable();
		DefaultComboBoxModel model = new DefaultComboBoxModel();
		this.userComboBoxModels.put("jabber.at", model);
		model = new DefaultComboBoxModel();
		this.userComboBoxModels.put("jabber.com", model);
		model = new DefaultComboBoxModel();
		this.userComboBoxModels.put("jabber.org", model);

		init();
	}

	public void init() {
		this.result = JOptionPane.CANCEL_OPTION;

		this.server = "";
		this.serverComboBox = new JComboBox(this.serverComboBoxModel);
		this.serverComboBox.setEditable(true);
		final ChatDialog thisChatDialog = this;
		this.serverComboBox.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent e) {
				if(thisChatDialog.serverComboBox.getSelectedIndex() != -1) {
					String item = (String)thisChatDialog.serverComboBox.getSelectedItem();
					ComboBoxModel model = (ComboBoxModel)thisChatDialog.userComboBoxModels.get(item);
					thisChatDialog.userComboBox.setModel(model);
				} else {
					ComboBoxModel model = new DefaultComboBoxModel();
					thisChatDialog.userComboBox.setModel(model);
					thisChatDialog.userComboBoxModels.put(thisChatDialog.serverComboBox.getSelectedItem(), model);
				}
			}
		});

		this.user = "";
		this.userComboBox = new JComboBox();
		this.userComboBox.setEditable(true);

		if(this.serverComboBoxModel.getSize() > 0) {
			this.serverComboBox.setSelectedIndex(0);

			ComboBoxModel userModel = (ComboBoxModel)this.userComboBoxModels.get(this.serverComboBox.getSelectedItem());
			this.userComboBox.setModel(userModel);
			if(userModel.getSize() > 0)
				this.userComboBox.setSelectedIndex(0);
		}

		this.subjectTextField = new JTextField();

		JLabel serverLabel = new JLabel("Server");
		JLabel userLabel = new JLabel("User");
		JLabel subjectLabel = new JLabel("Subject");

		Action okAction = new AbstractAction("Ok") {
			public void actionPerformed(ActionEvent e) {
				thisChatDialog.doAction();
			}
		};

		Action cancelAction = new AbstractAction("Cancel") {
			public void actionPerformed(ActionEvent e) {
				thisChatDialog.hide();
			}
		};

		JButton okButton = new JButton(okAction);
		okButton.setDefaultCapable(true);

		JButton cancelButton = new JButton(cancelAction);

		super.getContentPane().setLayout(new GridBagLayout());
		GridBagConstraints c = new GridBagConstraints();

		c.gridwidth = 2;
		c.gridy = 0;
		c.anchor = GridBagConstraints.SOUTHWEST;
		c.weighty = 1.0;
		c.weightx = 1.0;
		c.insets = new Insets(3, 3, 3, 3);
		super.getContentPane().add(serverLabel, c);

		c.gridy = 1;
		c.anchor = GridBagConstraints.WEST;
		c.weighty = 0.0;
		c.fill = GridBagConstraints.HORIZONTAL;
		super.getContentPane().add(this.serverComboBox, c);

		c.gridy = 2;
		c.anchor = GridBagConstraints.SOUTHWEST;
		c.weighty = 1.0;
		c.fill = GridBagConstraints.NONE;
		super.getContentPane().add(userLabel, c);

		c.gridy = 3;
		c.anchor = GridBagConstraints.WEST;
		c.weighty = 0.0;
		c.fill = GridBagConstraints.HORIZONTAL;
		super.getContentPane().add(this.userComboBox, c);

		c.gridy = 4;
		c.anchor = GridBagConstraints.WEST;
		c.weighty = 0.0;
		c.fill = GridBagConstraints.HORIZONTAL;
		super.getContentPane().add(subjectLabel, c);

		c.gridy = 5;
		c.anchor = GridBagConstraints.WEST;
		c.weighty = 0.0;
		c.fill = GridBagConstraints.HORIZONTAL;
		super.getContentPane().add(this.subjectTextField, c);

		c.gridwidth = 1;
		c.gridy = 6;
		c.anchor = GridBagConstraints.EAST;
		c.weighty = 1.0;
		c.fill = GridBagConstraints.NONE;
		super.getContentPane().add(okButton, c);

		c.gridx = 1;
		c.anchor = GridBagConstraints.WEST;
		super.getContentPane().add(cancelButton, c);

		super.getRootPane().setDefaultButton(okButton);

		super.enableEvents(AWTEvent.KEY_EVENT_MASK);
	}

	public void show() {
		this.result = JOptionPane.CANCEL_OPTION;
		this.subjectTextField.setText("{no subject}");
		this.subjectTextField.setSelectionStart(0);
		this.subjectTextField.setSelectionEnd(this.subjectTextField.getText().length());
		super.pack();
		super.setLocationRelativeTo(super.getParent());
		super.show();
	}

	public int getResult() {
		return this.result;
	}

	public String getServer() {
		return this.server;
	}

	public String getUser() {
		return this.user;
	}

	public String getSubject() {
		return this.subjectTextField.getText();
	}

	protected void processKeyEvent(KeyEvent e) {
		if(e.getKeyCode() == KeyEvent.VK_ESCAPE)
			this.hide();
	}

	public void doAction() {
		this.server = (String)this.serverComboBox.getSelectedItem();
		this.user = (String)this.userComboBox.getSelectedItem();

		MutableComboBoxModel serverModel = (MutableComboBoxModel)this.serverComboBox.getModel();
		MutableComboBoxModel userModel = (MutableComboBoxModel)this.userComboBox.getModel();

		if(this.serverComboBox.getSelectedIndex() != -1)
			serverModel.removeElement(this.server);

		serverModel.insertElementAt(this.server, 0);
		this.serverComboBox.setSelectedIndex(0);

		if(this.userComboBox.getSelectedIndex() != -1)
			userModel.removeElement(this.user);

		userModel.insertElementAt(this.user, 0);
		this.userComboBox.setSelectedIndex(0);

		this.result = JOptionPane.OK_OPTION;
		this.hide();
	}

	public void writeExternal(ObjectOutput oo) throws IOException {
		oo.writeInt(this.serverComboBoxModel.getSize());
		for(int i = 0; i < this.serverComboBoxModel.getSize(); i++) {
			String str = this.serverComboBoxModel.getElementAt(i).toString();
			oo.writeUTF(str);

			DefaultComboBoxModel model = (DefaultComboBoxModel)this.userComboBoxModels.get(str);
			oo.writeInt(model.getSize());
			for(int j = 0; j < model.getSize(); j++)
				oo.writeUTF(model.getElementAt(j).toString());
		}
		oo.flush();
	}

	public void readExternal(ObjectInput oi) throws IOException, ClassNotFoundException {
		this.serverComboBoxModel = new DefaultComboBoxModel();
		this.userComboBoxModels = new Hashtable();

		int serverCount = oi.readInt();
		for(int i = 0; i < serverCount; i++) {
			String str = oi.readUTF();
			this.serverComboBoxModel.addElement(str);

			DefaultComboBoxModel userComboBoxModel = new DefaultComboBoxModel();
			this.userComboBoxModels.put(str, userComboBoxModel);

			int userCount = oi.readInt();
			for(int j = 0; j < userCount; j++)
				userComboBoxModel.addElement(oi.readUTF());
		}

		init();
	}

}