package org.convey;

import javax.swing.*;

import java.awt.*;
import java.awt.event.*;

import java.util.*;

/**
 * @version 0.3
 */
public class ColorComboBox extends JComboBox implements ActionListener {

	private static Color previousColor = null;

	private static Vector colorNames = new Vector();

	private static Vector colorIcons = new Vector();

	private static ColorIcon noColorIcon = new ColorIcon(null);

	private Vector customColorIcons = new Vector();

	private int lastCustomOverwrite = -1;

	private int customCount;

	private boolean allowNone;

	static {
		colorNames.add("Black");
		colorNames.add("Silver");
		colorNames.add("Gray");
		colorNames.add("White");
		colorNames.add("Maroon");
		colorNames.add("Red");
		colorNames.add("Purple");
		colorNames.add("Fuchsia");
		colorNames.add("Green");
		colorNames.add("Lime");
		colorNames.add("Olive");
		colorNames.add("Yellow");
		colorNames.add("Navy");
		colorNames.add("Blue");
		colorNames.add("Teal");
		colorNames.add("Aqua");

		colorIcons.add(new ColorIcon(new Color(0, 0, 0)));
		colorIcons.add(new ColorIcon(new Color(192, 192, 192)));
		colorIcons.add(new ColorIcon(new Color(128, 128, 128)));
		colorIcons.add(new ColorIcon(new Color(255, 255, 255)));
		colorIcons.add(new ColorIcon(new Color(128, 0, 0)));
		colorIcons.add(new ColorIcon(new Color(255, 0, 0)));
		colorIcons.add(new ColorIcon(new Color(128, 0, 128)));
		colorIcons.add(new ColorIcon(new Color(255, 0, 255)));
		colorIcons.add(new ColorIcon(new Color(0, 192, 0)));
		colorIcons.add(new ColorIcon(new Color(0, 255, 0)));
		colorIcons.add(new ColorIcon(new Color(192, 192, 0)));
		colorIcons.add(new ColorIcon(new Color(255, 255, 0)));
		colorIcons.add(new ColorIcon(new Color(0, 0, 192)));
		colorIcons.add(new ColorIcon(new Color(0, 0, 255)));
		colorIcons.add(new ColorIcon(new Color(0, 192, 192)));
		colorIcons.add(new ColorIcon(new Color(0, 255, 255)));
	}

	private class ColorComboBoxRenderer extends JLabel implements ListCellRenderer {

		public ColorComboBoxRenderer() {
			setOpaque(true);
			super.setHorizontalAlignment(LEFT);
		}

		/*
		 * This method finds the image and text corresponding
		 * to the selected value and returns the label, set up
		 * to display the text and image.
		 */
		public Component getListCellRendererComponent(JList list, Object value,
																									int index, boolean isSelected,
																									boolean cellHasFocus) {

			//Get the selected index. (The index param isn't
			//always valid, so just use the value.)
			int selectedIndex = ((Integer)value).intValue();

			if(isSelected) {
				this.setBackground(list.getSelectionBackground());
				this.setForeground(list.getSelectionForeground());
			} else {
				this.setBackground(list.getBackground());
				this.setForeground(list.getForeground());
			}

			if(allowNone) {
				if(selectedIndex == 0) {
					super.setIcon(noColorIcon);
					super.setText("None");
				} else if(selectedIndex < colorNames.size() + 1) {
					super.setIcon((Icon)colorIcons.elementAt(selectedIndex - 1));
					super.setText((String)colorNames.elementAt(selectedIndex - 1));
				} else if(selectedIndex == colorNames.size() + 1) {
					super.setIcon(null);
					super.setText("Browse...");
				} else if(selectedIndex > colorNames.size() + 1) {
					super.setIcon((Icon)customColorIcons.elementAt(selectedIndex - colorNames.size() - 2));
					super.setText("Custom");
				}
			} else {
				if(selectedIndex < colorNames.size()) {
					super.setIcon((Icon)colorIcons.elementAt(selectedIndex));
					super.setText((String)colorNames.elementAt(selectedIndex));
				} else if(selectedIndex == colorNames.size()) {
					super.setIcon(null);
					super.setText("Browse...");
				} else if(selectedIndex > colorNames.size()) {
					super.setIcon((Icon)customColorIcons.elementAt(selectedIndex - colorNames.size() - 1));
					super.setText("Custom");
				}
			}

			return this;
		}

	}

	public ColorComboBox() {
		this(8, true);
	}

	public ColorComboBox(boolean allowNone) {
		this(8, allowNone);
	}

	public ColorComboBox(int customCount) {
		this(customCount, true);
	}

	public ColorComboBox(int customCount, boolean allowNone) {
		this.allowNone = allowNone;

		Integer[] intArray = null;
		if(customCount <= 0) {
			this.customCount = 0;

			if(this.allowNone) {
				intArray = new Integer[colorNames.size() + 1];
				super.setMaximumRowCount(colorNames.size() + 1);
			} else {
				intArray = new Integer[colorNames.size()];
				super.setMaximumRowCount(colorNames.size());
			}

			for(int i = 0; i < intArray.length; i++)
				intArray[i] = new Integer(i);
		} else {
			this.customCount = customCount;

			if(this.allowNone) {
				intArray = new Integer[colorNames.size() + 2];
				super.setMaximumRowCount(colorNames.size() + 2 + this.customCount);
			} else {
				intArray = new Integer[colorNames.size() + 1];
				super.setMaximumRowCount(colorNames.size() + 1 + this.customCount);
			}

			for(int i = 0; i < intArray.length; i++)
				intArray[i] = new Integer(i);
		}

		super.setModel(new DefaultComboBoxModel(intArray));
		super.setRenderer(new ColorComboBoxRenderer());

		super.addActionListener(this);
	}

	public boolean isFocusTraversable() {
		return false;
	}

	public Color getSelectedColor() {
		int selectedIndex = ((Integer)super.getSelectedItem()).intValue();

		if(this.allowNone) {
			if(selectedIndex > 0 && selectedIndex < colorNames.size() + 1)
				return ((ColorIcon)colorIcons.elementAt(selectedIndex - 1)).getColor();
			else if(selectedIndex > colorNames.size() + 1)
				return ((ColorIcon)this.customColorIcons.elementAt(selectedIndex - colorNames.size() - 2)).getColor();
		} else {
			if(selectedIndex >= 0 && selectedIndex < colorNames.size())
				return ((ColorIcon)colorIcons.elementAt(selectedIndex)).getColor();
			else if(selectedIndex > colorNames.size())
				return ((ColorIcon)this.customColorIcons.elementAt(selectedIndex - colorNames.size() - 1)).getColor();
		}

		return null;
	}

	public void setSelectedColor(Color color) {
		if(color == null) {
			if(this.allowNone)
				this.setSelectedIndex(0);
			else
				return;
		} else {
			if(this.allowNone) {
				for(int i = 0; i < this.colorIcons.size(); i++) {
					if(((ColorIcon)this.colorIcons.elementAt(i)).getColor().equals(color)) {
						this.setSelectedIndex(i + 1);
						return;
					}
				}
			} else {
				for(int i = 0; i < this.colorIcons.size(); i++) {
					if(((ColorIcon)this.colorIcons.elementAt(i)).getColor().equals(color)) {
						this.setSelectedIndex(i);
						return;
					}
				}
			}
		}

		if(this.customCount > 0) {
			if(this.allowNone) {
				for(int i = 0; i < this.customColorIcons.size(); i++) {
					if(((ColorIcon)this.customColorIcons.elementAt(i)).getColor().equals(color)) {
						this.setSelectedIndex(i + colorNames.size() + 2);
						return;
					}
				}

				if(this.customColorIcons.size() < this.customCount) {
					this.customColorIcons.add(new ColorIcon(color));
					((DefaultComboBoxModel)this.getModel()).addElement(new Integer(this.customColorIcons.size() + colorNames.size() + 1));
					this.setSelectedIndex(this.customColorIcons.size() + colorNames.size() + 1);
				} else {
					if(this.lastCustomOverwrite == this.customCount - 1)
						this.lastCustomOverwrite = 0;
					else
						this.lastCustomOverwrite++;
					this.customColorIcons.removeElementAt(lastCustomOverwrite);
					this.customColorIcons.insertElementAt(new ColorIcon(color), lastCustomOverwrite);
					this.setSelectedIndex(colorNames.size() + 2 + lastCustomOverwrite);
				}
			} else {
				for(int i = 0; i < this.customColorIcons.size(); i++) {
					if(((ColorIcon)this.customColorIcons.elementAt(i)).getColor().equals(color)) {
						this.setSelectedIndex(i + colorNames.size() + 1);
						return;
					}
				}

				if(this.customColorIcons.size() < this.customCount) {
					this.customColorIcons.add(new ColorIcon(color));
					((DefaultComboBoxModel)this.getModel()).addElement(new Integer(this.customColorIcons.size() + colorNames.size()));
					this.setSelectedIndex(this.customColorIcons.size() + colorNames.size());
				} else {
					if(this.lastCustomOverwrite == this.customCount - 1)
						this.lastCustomOverwrite = 0;
					else
						this.lastCustomOverwrite++;
					this.customColorIcons.removeElementAt(lastCustomOverwrite);
					this.customColorIcons.insertElementAt(new ColorIcon(color), lastCustomOverwrite);
					this.setSelectedIndex(colorNames.size() + 1 + lastCustomOverwrite);
				}
			}
		}
	}

	public void actionPerformed(ActionEvent e) {
		if(this.getSelectedIndex() == colorNames.size() + 1 && this.allowNone ||
			 this.getSelectedIndex() == colorNames.size() && !this.allowNone) {
			Color newColor = JColorChooser.showDialog(this, "Color", this.previousColor);
			if(newColor == null)
				this.setSelectedColor(this.previousColor);
			else
				this.setSelectedColor(newColor);

			this.previousColor = this.getSelectedColor();

			ActionListener[] als = (ActionListener[])super.listenerList.getListeners(ActionListener.class);
			for(int i = 0; i < als.length; i++) {
				if(als[i] != this)
					als[i].actionPerformed(e);
			}
		}
	}

}

class ColorIcon implements Icon {

	private Color color;

	public ColorIcon(Color color) {
		this.color = color;
	}

	public int getIconHeight() {
		return 16;
	}

	public int getIconWidth() {
		return 16;
	}

	public void paintIcon(Component c, Graphics g, int x, int y) {
		if (this.color != null) {
			g.setColor(this.color);
			g.fillRect(x, y, x + 16, y + 16);

			g.setColor(Color.black);
			g.drawRect(x, y, x + 15, y + 15);
		}
		else {
			g.setColor(Color.red);
			g.drawLine(x, y, x + 15, y + 15);
			g.drawLine(x, y + 15, x + 15, y);

			g.setColor(Color.black);
			g.drawRect(x, y, x + 15, y + 15);
		}
	}

	public Color getColor() {
		return this.color;
	}

}