package org.convey;

import java.awt.*;

import java.awt.geom.*;

import java.awt.event.*;

import org.jdom.*;

/**
 * @author Andy Ames
 * @version 0.2
 */
public class EllipseGraphic extends ShapeGraphic {

	protected Rectangle ellipse;

	protected transient Rectangle origEllipse;

	protected transient int vertexMoving = -1;

	public EllipseGraphic() {
		this.name = "ellipse1";
	}

	public EllipseGraphic(GraphicPanel graphicPanel) {
		super(graphicPanel);
		this.name = "ellipse1";
	}

	/**
	 * Constructor.
	 */
	public EllipseGraphic(GraphicPanel graphicPanel, Point p) {
		super(graphicPanel);
		ellipse = new Rectangle(p.x, p.y, 0, 0);
		updateCache();
		super.repaint();
		this.name = "ellipse1";
	}

	public Element toElement() {
		Element element = super.toElement();
		element.setName("ellipse");

		int halfWidth = (int)((double)ellipse.width / 2.0);
		int halfHeight = (int)((double)ellipse.height / 2.0);

		element.setAttribute("cx", String.valueOf(ellipse.x + halfWidth));
		element.setAttribute("cy", String.valueOf(ellipse.y + halfHeight));
		element.setAttribute("rx", String.valueOf(halfWidth));
		element.setAttribute("ry", String.valueOf(halfHeight));

		return element;
	}

	public void fromElement(Element element) {
		super.fromElement(element);

		int cx = Integer.parseInt(element.getAttributeValue("cx"));
		int cy = Integer.parseInt(element.getAttributeValue("cy"));
		int rx = Integer.parseInt(element.getAttributeValue("rx"));
		int ry = Integer.parseInt(element.getAttributeValue("ry"));

		this.ellipse = new Rectangle(cx - rx, cy - ry, 2 * rx, 2 * ry);
		this.updateCache();
		this.repaint();
	}

	public void paintSelection(Graphics2D g2d) {
		switch(super.graphicPanel.getState()) {
			case GraphicPanel.SCALE:
				super.paintSelection(g2d);
				break;
			default:
				g2d.setStroke(new BasicStroke());

				g2d.setColor(Color.black);

				g2d.fillRect(this.ellipse.x - 2, this.ellipse.y + this.ellipse.height / 2 - 2, 5, 5);
				g2d.fillRect(this.ellipse.x + this.ellipse.width - 2, this.ellipse.y + this.ellipse.height / 2 - 2, 5, 5);

				g2d.fillRect(this.ellipse.x + this.ellipse.width / 2 - 2, this.ellipse.y - 2, 5, 5);
				g2d.fillRect(this.ellipse.x + this.ellipse.width / 2 - 2, this.ellipse.y + this.ellipse.height - 2, 5, 5);
				break;
		}
	}

	public boolean hitVertex(Point p) {
		if(super.getState() == Graphic.CREATING)
			return false;
		switch(this.graphicPanel.getState()) {
			case GraphicPanel.SCALE:
				return super.hitVertex(p);
			default:
				int xcoin = -1; // 0 is left, 1 is right, 2 is center
				int ycoin = -1; // 0 is top, 1 is bottom, 2 is middle

				if(p.x >= this.bounds.x - 3 && p.x <= this.bounds.x + 3) {
					// left edge
					xcoin = 0;
				} else if(p.x >= this.bounds.x + this.bounds.width - 3 && p.x <= this.bounds.x + this.bounds.width + 3) {
					// right edge
					xcoin = 1;
				}  else if(p.x >= this.bounds.x + this.bounds.width / 2 - 3 && p.x <= this.bounds.x + this.bounds.width / 2 + 3) {
					// center
					xcoin = 2;
				}

				if(xcoin == -1)
					return false;

				if(p.y >= this.bounds.y - 3 && p.y <= this.bounds.y + 3) {
					// top edge
					ycoin = 0;
				} else if(p.y >= this.bounds.y + this.bounds.height - 3 && p.y <= this.bounds.y + this.bounds.height + 3) {
					// bottom edge
					ycoin = 1;
				}  else if(p.y >= this.bounds.y + this.bounds.height / 2 - 3 && p.y <= this.bounds.y + this.bounds.height / 2 + 3) {
					// middle
					ycoin = 2;
				}

				if(ycoin == 2 && (xcoin == 0 || xcoin == 1))
					return true;
				if(xcoin == 2 && (ycoin == 0 || ycoin == 1))
					return true;

				break;
		}

		return false;
	}

	public void fitToRect(Rectangle r) {
		this.ellipse = (Rectangle)r.clone();

		super.fitToRect(r);
	}

	/**
	 * @todo: Implement vertex manipulation
	 */
	public void mousePressed(MouseEvent e) {
		if(super.getState() == Graphic.CREATING)
			return;
		switch(super.graphicPanel.getState()) {
			case GraphicPanel.SCALE:
				super.mousePressed(e);
				break;
			default:
				Point p = e.getPoint();
				int xcoin = -1; // -1 is not coincident, 0 is left, 1 is right, 2 is center
				int ycoin = -1; // -1 is not coincident, 0 is top, 1 is bottom, 2 is middle

				if(p.x >= this.bounds.x - 3 && p.x <= this.bounds.x + 3) {
					// left edge
					xcoin = 0;
				} else if(p.x >= this.bounds.x + this.bounds.width - 3 && p.x <= this.bounds.x + this.bounds.width + 3) {
					// right edge
					xcoin = 1;
				}  else if(p.x >= this.bounds.x + this.bounds.width / 2 - 3 && p.x <= this.bounds.x + this.bounds.width / 2 + 3) {
					// center
					xcoin = 2;
				}

				if(p.y >= this.bounds.y - 3 && p.y <= this.bounds.y + 3) {
					// top edge
					ycoin = 0;
				} else if(p.y >= this.bounds.y + this.bounds.height - 3 && p.y <= this.bounds.y + this.bounds.height + 3) {
					// bottom edge
					ycoin = 1;
				}  else if(p.y >= this.bounds.y + this.bounds.height / 2 - 3 && p.y <= this.bounds.y + this.bounds.height / 2 + 3) {
					// middle
					ycoin = 2;
				}

				scaleAnchor = -1;
				if(xcoin == 0) {
					if(ycoin == 2) {
						this.scaleAnchor = RIGHT;
					}
				} else if(xcoin == 1) {
					if(ycoin == 2) {
						this.scaleAnchor = LEFT;
					}
				} else if(xcoin == 2) {
					if(ycoin == 0) {
						this.scaleAnchor = BOTTOM;
					} else if(ycoin == 1) {
						this.scaleAnchor = TOP;
					}
				}
				break;
		}

		lastDragPoint = pressPoint = e.getPoint();
	}

	public void setOrig() {
		super.setOrig();
		this.origEllipse = (Rectangle)this.ellipse.clone();
	}

	/**
	 * Invoked when a mouse button has been released on a component.
	 */
	public void mouseReleased(MouseEvent e) {
		if(super.getState() == CREATING) {
			Point p = e.getPoint();
			ellipse.width = p.x - ellipse.x;
			ellipse.height = p.y - ellipse.y;
			super.setState(DORMANT);
		} else if(this.scaleAnchor != -1) {
			if(super.lastFitToRect != null) {
				this.graphicPanel.pushGraphicEvent(new FitToRectGraphicEvent(this, this.origBounds, this.lastFitToRect));
				this.scaleAnchor = -1;
				super.lastFitToRect = null;
			}
		} else
			super.mouseReleased(e);
	}

/**
	* Invoked when a mouse button is pressed on a component and then
	* dragged.  Mouse drag events will continue to be delivered to
	* the component where the first originated until the mouse button is
	* released (regardless of whether the mouse position is within the
	* bounds of the component).
	*/
	public void mouseDragged(MouseEvent e) {
		if(super.getState() == CREATING) {
			Point p = e.getPoint();
			ellipse.width = p.x - ellipse.x;
			ellipse.height = p.y - ellipse.y;

			updateCache();
			super.repaint();
		} else {
			Point p = e.getPoint();
			switch(super.graphicPanel.getState()) {
				case GraphicPanel.SCALE:
					super.mouseDragged(e);
					break;
				default:
					if(scaleAnchor != -1) {
						this.lastFitToRect = (Rectangle)this.origBounds.clone();
						switch(scaleAnchor) {
							case TOP_LEFT:
								this.lastFitToRect.x = this.origBounds.x;
								this.lastFitToRect.width = p.x - this.origBounds.x;

								this.lastFitToRect.y = this.origBounds.y;
								this.lastFitToRect.height = p.y - this.origBounds.y;
								break;
							case TOP:
								this.lastFitToRect.y = this.origBounds.y;
								this.lastFitToRect.height = p.y - this.origBounds.y;
								break;
							case TOP_RIGHT:
								this.lastFitToRect.x = p.x;
								this.lastFitToRect.width = this.origBounds.x + this.origBounds.width - p.x;

								this.lastFitToRect.y = this.origBounds.y;
								this.lastFitToRect.height = p.y - this.origBounds.y;
								break;
							case RIGHT:
								this.lastFitToRect.x = p.x;
								this.lastFitToRect.width = this.origBounds.x + this.origBounds.width - p.x;
								break;
							case BOTTOM_RIGHT:
								this.lastFitToRect.x = p.x;
								this.lastFitToRect.width = this.origBounds.x + this.origBounds.width - p.x;

								this.lastFitToRect.y = p.y;
								this.lastFitToRect.height = this.origBounds.y + this.origBounds.height - p.y;
								break;
							case BOTTOM:
								this.lastFitToRect.y = p.y;
								this.lastFitToRect.height = this.origBounds.y + this.origBounds.height - p.y;
								break;
							case BOTTOM_LEFT:
								this.lastFitToRect.x = this.origBounds.x;
								this.lastFitToRect.width = p.x - this.origBounds.x;

								this.lastFitToRect.y = p.y;
								this.lastFitToRect.height = this.origBounds.y + this.origBounds.height - p.y;
								break;
							case LEFT:
								this.lastFitToRect.x = this.origBounds.x;
								this.lastFitToRect.width = p.x - this.origBounds.x;
								break;
						}
						this.fitToRect(this.lastFitToRect);
						this.updateCache();
						super.repaint();
					} else {
						this.translate(p.x - lastDragPoint.x, p.y - lastDragPoint.y);
						lastDragPoint = p;
					}
					break;
			}
		}
	}

	public void updateCache() {
		float x = Math.min(ellipse.x, ellipse.x + ellipse.width);
		float y = Math.min(ellipse.y, ellipse.y + ellipse.height);
		float w = Math.abs(ellipse.width);
		float h = Math.abs(ellipse.height);
		super.shape = new Ellipse2D.Float(x, y, w, h);
		Area hitArea = new Area(new BasicStroke(6).createStrokedShape(super.shape));
		if(super.fillColor != null)
			hitArea.add(new Area(super.shape));
		super.pathHotSpot = hitArea;
		super.bounds = new Rectangle((int)x, (int)y, (int)w, (int)h);
		super.currDirty = new Rectangle((int)x, (int)y, (int)w, (int)h);
		super.currDirty.grow(5, 5);
	}

	public String getStatus() {
		return "ELLIPSE: (" + ellipse.x + "," + ellipse.y + "," + ellipse.width + "," + ellipse.height + ")";
	}

	public void translate(int x, int y) {
		super.translate(x, y);
		ellipse.translate(x, y);
		super.repaint();
	}

	public void setEllipse(Rectangle ellipse) {
		this.ellipse = (Rectangle)ellipse.clone();
		this.updateCache();
		super.repaint();
	}

	public void undoGraphicEvent(GraphicEvent event) {
		super.undoGraphicEvent(event);

		if(event instanceof EllipseGraphicEvent) {
			EllipseGraphicEvent ellipseGraphicEvent = (EllipseGraphicEvent)event;

			this.setEllipse(ellipseGraphicEvent.getOrigEllipse());
		}
	}

	public void redoGraphicEvent(GraphicEvent event) {
		super.redoGraphicEvent(event);

		if(event instanceof EllipseGraphicEvent) {
			EllipseGraphicEvent ellipseGraphicEvent = (EllipseGraphicEvent)event;

			this.setEllipse(ellipseGraphicEvent.getNewEllipse());
		}
	}

}