package org.convey;

import java.awt.*;
import java.awt.event.*;
import java.util.*;
import java.awt.geom.*;

import org.jdom.*;

/**
 * @author Ed Braunhut
 * @version 0.1
 */
public class FreehandGraphic extends ShapeGraphic {

	private transient Vector points = new Vector();

	protected Vector origPoints = new Vector();

	public FreehandGraphic() {
		this.name = "freehand1";

	}

	public FreehandGraphic(GraphicPanel graphicPanel) {
		super(graphicPanel);
		this.name = "freehand1";
	}

	public FreehandGraphic(GraphicPanel graphicPanel, Point p) {
		super(graphicPanel);
		points.add(new Point2D.Double(p.x, p.y));
		updateCache();
		super.repaint();
		this.name = "freehand1";
	}
	/**
	 * <polyline fill="none" stroke="blue"
	 *    points="50,375
	 *            150,375 150,325 250,325 250,375
	 *            350,375 350,250 450,250 450,375
	 *            550,375 550,175 650,175 650,375
	 *            750,375 750,100 850,100 850,375
	 *            950,375 950,25 1050,25 1050,375
	 *            1150,375" />
	 *
	 * <path d="M 100 100 L 300 100 L 200 300 z"
	 *    fill="red" stroke="blue" />
	 */
	public Element toElement() {
		Element element = super.toElement();
		element.setName( "path" );

		StringBuffer buffer = new StringBuffer();
		buffer.append("M ");
		Iterator iter = points.iterator();
		while (iter.hasNext()) {
			Point2D p = (Point2D)iter.next();
			buffer.append(p.getX());
			buffer.append(" ");
			buffer.append(p.getY());
			buffer.append(" L ");
		}
		if (buffer.length() > 0) {
			buffer.delete(buffer.length() - 3, buffer.length() - 1);
		}
		element.setAttribute( "d" , buffer.toString());
		return element;
	}

	public void fromElement(Element element) {
		super.fromElement(element);
		StringTokenizer tokenizer = new StringTokenizer(element.getAttributeValue( "d" ), " ML" );
		this.points.clear();
		while (tokenizer.hasMoreTokens()) {
			double x = Double.parseDouble(tokenizer.nextToken());
			double y = Double.parseDouble(tokenizer.nextToken());
			this.points.add(new Point2D.Double(x, y));
		}
		this.updateCache();
		this.repaint();
	}

	public void paintSelection(Graphics2D g2d) {
		switch(super.graphicPanel.getState()) {
			case GraphicPanel.SCALE:
				super.paintSelection(g2d);
				break;
			default:
				g2d.setColor(Color.black);
				g2d.setStroke(new BasicStroke());
				break;
			}
	}

	/**
	 * Invoked when a mouse button has been released on a component. If the
	 * preceding mouseDragged method collected less than 4 points for the points
	 * vector, the points vector is set to null as there is not enough points to
	 * use the curve fitting algorithm. Otherwise a new cubic curve graphic is
	 * created by calling the createFittedCurve method.
	 */
	public void mouseReleased(MouseEvent e) {
		if (super.getState() == CREATING) {

				super.setState(DORMANT);
		}
		super.mouseReleased(e);
	}

	/**
	* Invoked when a mouse button is pressed on a component and then
	* dragged.  Mouse drag events will continue to be delivered to
	* the component where the first originated until the mouse button is
	* released (regardless of whether the mouse position is within the
	* bounds of the component).
	*/
	public void mouseDragged(MouseEvent e) {
		if(super.getState() == CREATING) {
			Point p = e.getPoint();
			points.add(new Point2D.Double(p.x, p.y));
			updateCache();
			super.repaint();
		} else
			super.mouseDragged(e);
	}

	public void setOrig() {
		super.setOrig();

		this.origPoints.clear();
		Iterator iter = this.points.iterator();
		while(iter.hasNext())
			this.origPoints.add(((Point2D)iter.next()).clone());
	}

	/**
	 * Method getStatus
	 *
	 * @return   a String
	 *
	 */
	public String getStatus() {
		this.name = "freehand";
		return "";
	}

	public void fitToRect(Rectangle r) {
		if(this.origBounds.width == 0 || this.origBounds.height == 0)
			return;
		double tx1 = (double)-super.origBounds.x;
		double ty1 = (double)-super.origBounds.y;
		double sx = (double)r.width / (double)super.origBounds.width;
		double sy = (double)r.height / (double)super.origBounds.height;
		double tx2 = (double)r.x;
		double ty2 = (double)r.y;

		Iterator origIter = this.origPoints.iterator();
		Iterator iter = this.points.iterator();
		while(iter.hasNext()) {
			Point2D origP = (Point2D)origIter.next();
			Point2D p = (Point2D)iter.next();
			p.setLocation((origP.getX() + tx1) * sx + tx2, (origP.getY() + ty1) * sy + ty2);
		}

		super.fitToRect(r);
	}

	/**
	 *
	 */
	public void updateCache() {
		GeneralPath path = new GeneralPath();
		Point2D p = (Point2D)points.elementAt(0);
		path.moveTo((float)p.getX(), (float)p.getY());
		for(int i = 1; i < points.size(); i++) {
			p = (Point2D)points.elementAt(i);
			path.lineTo((float)p.getX(), (float)p.getY());
		}
		super.shape = path;
		Area area = new Area(new BasicStroke(6).createStrokedShape(super.shape));
		if(super.fillColor != null)
			area.add(new Area(super.shape));
		super.pathHotSpot = area;
		super.currDirty = super.shape.getBounds();
		super.currDirty.grow(5,5);
		super.bounds = super.shape.getBounds();
	}

	public void translate(int x, int y) {
		super.translate(x, y);
		Iterator iter = points.iterator();
		while(iter.hasNext()) {
			Point2D p = (Point2D)iter.next();
			p.setLocation(p.getX() + (double)x, p.getY() + (double)y);
		}
		super.repaint();
	}

	public void setPoints(Vector points) {
		this.points = points;
		this.updateCache();
		super.repaint();
	}

}