package org.convey;
import java.awt.Graphics2D;
import java.awt.event.MouseEvent;
import java.awt.event.KeyEvent;
import java.util.*;
import java.awt.*;
import java.awt.geom.*;

import org.jdom.*;

/**
 * GroupGraphic.java
 *
 * @author Ed Braunhut
 * @version 0.2
 */
public class GroupGraphic extends Graphic {

	protected Vector group = new Vector();

	public GroupGraphic(GraphicPanel graphicPanel){
		super(graphicPanel);
		this.name = "group1";
	}

	/**
	 * Paints the Graphic object onto the Graphics2D object.
	 */
	public void paint(Graphics2D g2d) {
		Iterator iter = group.iterator();
		while(iter.hasNext()) {
			Graphic graphic = (Graphic)iter.next();
			graphic.paint(g2d);
		}
	}

	public void paintSelection(Graphics2D g2d) {
		switch(super.graphicPanel.getState()) {
			case GraphicPanel.SCALE:
				super.paintSelection(g2d);
				break;
			default:
				g2d.setStroke(new BasicStroke());
				g2d.setColor(Color.gray);

				g2d.draw(this.bounds);

				g2d.setColor(Color.black);

				g2d.fillRect(super.bounds.x - 2, super.bounds.y - 2, 5, 5);
				g2d.fillRect(super.bounds.x + super.bounds.width - 2, super.bounds.y - 2, 5, 5);
				g2d.fillRect(super.bounds.x + super.bounds.width - 2, super.bounds.y + super.bounds.height - 2, 5, 5);
				g2d.fillRect(super.bounds.x - 2 , super.bounds.y + super.bounds.height - 2, 5, 5);
				break;
		}
	}

	public Element toElement() {
		Element element = super.toElement();

		element.setName("g");
		Iterator iter = this.iterator();
		while(iter.hasNext()) {
			Graphic graphic = (Graphic)iter.next();
			element.addContent(graphic.toElement());
		}

		return element;
	}

	public void fromElement(Element element) {
		super.fromElement(element);

		Iterator iter = element.getChildren().iterator();
		while(iter.hasNext())
			this.addGraphic(this.graphicPanel.createGraphicFromSVGElement((Element)iter.next()));
		this.updateCache();
		this.repaint();
	}

	public boolean hitVertex(Point p) {
		if(super.getState() == Graphic.CREATING)
			return false;
		switch(this.graphicPanel.getState()) {
			case GraphicPanel.SCALE:
				return super.hitVertex(p);
			default:
				int xcoin = -1; // 0 is left, 1 is right, 2 is center
				int ycoin = -1; // 0 is top, 1 is bottom, 2 is middle

				if(p.x >= this.bounds.x - 3 && p.x <= this.bounds.x + 3) {
					// left edge
					xcoin = 0;
				} else if(p.x >= this.bounds.x + this.bounds.width - 3 && p.x <= this.bounds.x + this.bounds.width + 3) {
					// right edge
					xcoin = 1;
				}  else if(p.x >= this.bounds.x + this.bounds.width / 2 - 3 && p.x <= this.bounds.x + this.bounds.width / 2 + 3) {
					// center
					xcoin = 2;
				}

				if(xcoin == -1)
					return false;

				if(p.y >= this.bounds.y - 3 && p.y <= this.bounds.y + 3) {
					// top edge
					ycoin = 0;
				} else if(p.y >= this.bounds.y + this.bounds.height - 3 && p.y <= this.bounds.y + this.bounds.height + 3) {
					// bottom edge
					ycoin = 1;
				}  else if(p.y >= this.bounds.y + this.bounds.height / 2 - 3 && p.y <= this.bounds.y + this.bounds.height / 2 + 3) {
					// middle
					ycoin = 2;
				}

				if((xcoin == 0 || xcoin == 1) && (ycoin == 0 || ycoin == 1))
					return true;

				break;
		}

		return false;
	}

	public String getStatus() {
		return "grouping selected graphics...";
	}

	public void mousePressed(MouseEvent e) {
		switch(super.graphicPanel.getState()) {
			case GraphicPanel.SCALE:
				super.mousePressed(e);
				break;
			default:
				Iterator iter = this.group.iterator();
				while(iter.hasNext()) {
					Graphic g = (Graphic)iter.next();
					g.mousePressed(e);
				}
				Point p = e.getPoint();
				int xcoin = -1; // -1 is not coincident, 0 is left, 1 is right
				int ycoin = -1; // -1 is not coincident, 0 is top, 1 is bottom

				if(p.x >= this.bounds.x - 3 && p.x <= this.bounds.x + 3) {
					// left edge
					xcoin = 0;
				} else if(p.x >= this.bounds.x + this.bounds.width - 3 && p.x <= this.bounds.x + this.bounds.width + 3) {
					// right edge
					xcoin = 1;
				}

				if(p.y >= this.bounds.y - 3 && p.y <= this.bounds.y + 3) {
					// top edge
					ycoin = 0;
				} else if(p.y >= this.bounds.y + this.bounds.height - 3 && p.y <= this.bounds.y + this.bounds.height + 3) {
					// bottom edge
					ycoin = 1;
				}

				scaleAnchor = -1;
				if(xcoin == 0) {
					if(ycoin == 0)
						super.scaleAnchor = BOTTOM_RIGHT;
					else if(ycoin == 1)
						super.scaleAnchor = TOP_RIGHT;
				} else if(xcoin == 1) {
					if(ycoin == 0)
						super.scaleAnchor = BOTTOM_LEFT;
					else if(ycoin == 1)
						super.scaleAnchor = TOP_LEFT;
				}
				break;
		}

		lastDragPoint = pressPoint = e.getPoint();
	}

	/**
	* Invoked when a mouse button is pressed on a component and then
	* dragged.  Mouse drag events will continue to be delivered to
	* the component where the first originated until the mouse button is
	* released (regardless of whether the mouse position is within the
	* bounds of the component).
	*/
	public void mouseDragged(MouseEvent e) {
		Point p = e.getPoint();
		switch(super.graphicPanel.getState()) {
			case GraphicPanel.SCALE:
				super.mouseDragged(e);
				break;
			default:
				super.lastFitToRect = (Rectangle)super.origBounds.clone();
				if(scaleAnchor != -1) {
					switch(scaleAnchor) {
						case TOP_LEFT:
							this.lastFitToRect.x = this.origBounds.x;
							this.lastFitToRect.width = p.x - this.origBounds.x;

							this.lastFitToRect.y = this.origBounds.y;
							this.lastFitToRect.height = p.y - this.origBounds.y;
							break;
						case TOP_RIGHT:
							this.lastFitToRect.x = p.x;
							this.lastFitToRect.width = this.origBounds.x + this.origBounds.width - p.x;

							this.lastFitToRect.y = this.origBounds.y;
							this.lastFitToRect.height = p.y - this.origBounds.y;
							break;
						case BOTTOM_RIGHT:
							this.lastFitToRect.x = p.x;
							this.lastFitToRect.width = this.origBounds.x + this.origBounds.width - p.x;

							this.lastFitToRect.y = p.y;
							this.lastFitToRect.height = this.origBounds.y + this.origBounds.height - p.y;
							break;
						case BOTTOM_LEFT:
							this.lastFitToRect.x = this.origBounds.x;
							this.lastFitToRect.width = p.x - this.origBounds.x;

							this.lastFitToRect.y = p.y;
							this.lastFitToRect.height = this.origBounds.y + this.origBounds.height - p.y;
							break;
					}
					this.fitToRect(this.lastFitToRect);
					this.updateCache();
					super.repaint();
				} else {
					this.translate(p.x - lastDragPoint.x, p.y - lastDragPoint.y);
					lastDragPoint = p;
				}
				break;
		}
	}

	public void setOrig() {
		super.setOrig();
		Iterator iter = this.group.iterator();
		while(iter.hasNext()) {
			Graphic g = (Graphic)iter.next();
			g.setOrig();
		}
	}

	public void fitToRect(Rectangle r) {
		if(this.origBounds.width == 0 || this.origBounds.height == 0)
			return;
		double tx1 = (double)-super.origBounds.x;
		double ty1 = (double)-super.origBounds.y;
		double sx = (double)r.width / (double)super.origBounds.width;
		double sy = (double)r.height / (double)super.origBounds.height;
		double tx2 = (double)r.x;
		double ty2 = (double)r.y;

		Iterator iter = this.group.iterator();
		while(iter.hasNext()) {
			Graphic g = (Graphic)iter.next();
			Rectangle oBounds = g.getOrigBounds();
			Rectangle nBounds = new Rectangle();

			double x1 = ((double)oBounds.x + tx1) * sx + tx2;
			double y1 = ((double)oBounds.y + ty1) * sy + ty2;
			double x2 = ((double)(oBounds.x + oBounds.width) + tx1) * sx + tx2;
			double y2 = ((double)(oBounds.y + oBounds.height) + ty1) * sy + ty2;

			nBounds.setBounds((int)x1, (int)y1, (int)(x2 - x1), (int)(y2 - y1));

			g.fitToRect(nBounds);
		}

		super.fitToRect(r);
	}

	/**
	 * Here, we want to update the cache of this Graphic. Not the
	 * children. Any method that alters the graphic must call this method.
	 */
	public void updateCache() {
		Graphic graphic = (Graphic)group.elementAt(0);

		graphic.updateCache();

		super.bounds = graphic.getBounds();
		Area area = new Area(graphic.pathHotSpot);
		for(int i = 1; i < group.size(); i++) {
			graphic = (Graphic)group.elementAt(i);

			graphic.updateCache();

			super.bounds = super.bounds.union(graphic.getBounds());
			area.add(new Area(graphic.pathHotSpot));
		}
		super.pathHotSpot = area;
		super.currDirty = (Rectangle)super.bounds.clone();
		super.currDirty.grow(5, 5);
	}

	public Iterator iterator () {
		return group.iterator();
	}

	public void addGraphic (Graphic graphic) {
		this.group.add(graphic);
	}

	/**
	 * Translate the group by translating each of it's members
	 */
	public void translate(int x, int y) {
		super.translate(x, y);
		Iterator iter = this.iterator();
		while(iter.hasNext()) {
			Graphic graphic = (Graphic)iter.next();
			graphic.translate(x,y);
		}
		super.repaint();
	}

	public void undoGraphicEvent(GraphicEvent event) {
		super.undoGraphicEvent(event);

		if(event instanceof GroupGraphicEvent) {
			GroupGraphicEvent groupGraphicEvent = (GroupGraphicEvent)event;

			Iterator iter = null;
			switch(groupGraphicEvent.getType()) {
				case GroupGraphicEvent.TYPE_GROUP:
					this.graphicPanel.removeGraphic(this);
					this.graphicPanel.removeSelectedGraphic(this);
					this.setState(Graphic.DORMANT);

					iter = this.iterator();
					while (iter.hasNext()) {
						Graphic graphic = (Graphic)iter.next();

						this.graphicPanel.addGraphic(graphic);
						this.graphicPanel.addSelectedGraphic(graphic);
						graphic.setState(Graphic.SELECTED);
					}
					break;
				case GroupGraphicEvent.TYPE_UNGROUP:
					iter = this.iterator();
					while(iter.hasNext()) {
						Graphic graphic = (Graphic)iter.next();

						this.graphicPanel.removeGraphic(graphic);
						this.graphicPanel.removeSelectedGraphic(graphic);
						graphic.setState(Graphic.DORMANT);
					}

					this.graphicPanel.addGraphic(this);
					this.graphicPanel.addSelectedGraphic(this);
					this.setState(Graphic.SELECTED);
					break;
			}
		}
	}

	public void redoGraphicEvent(GraphicEvent event) {
		super.redoGraphicEvent(event);

		if(event instanceof GroupGraphicEvent) {
			GroupGraphicEvent groupGraphicEvent = (GroupGraphicEvent)event;

			Iterator iter = null;
			switch(groupGraphicEvent.getType()) {
				case GroupGraphicEvent.TYPE_GROUP:
					iter = this.iterator();
					while (iter.hasNext()) {
						Graphic graphic = (Graphic)iter.next();

						this.graphicPanel.removeGraphic(graphic);
						this.graphicPanel.removeSelectedGraphic(graphic);
						graphic.setState(Graphic.DORMANT);
					}

					this.graphicPanel.addGraphic(this);
					this.graphicPanel.addSelectedGraphic(this);
					this.setState(Graphic.SELECTED);
					break;
				case GroupGraphicEvent.TYPE_UNGROUP:
					this.graphicPanel.removeGraphic(this);
					this.graphicPanel.removeSelectedGraphic(this);
					this.setState(Graphic.DORMANT);

					iter = this.iterator();
					while (iter.hasNext()) {
						Graphic graphic = (Graphic) iter.next();

						this.graphicPanel.addGraphic(graphic);
						this.graphicPanel.addSelectedGraphic(graphic);
						graphic.setState(Graphic.SELECTED);
					}
					break;
			}
		}
	}

}