package org.convey;

import javax.swing.*;
import javax.swing.text.*;

import java.awt.*;

import java.awt.event.*;

import com.echomine.jabber.*;
import com.echomine.common.*;
import com.echomine.net.*;

import org.jdom.*;

import java.util.*;

import java.text.*;

/**
 * @author Andy Ames
 * @version 0.3
 */
public class JabberCanvasPanel extends CanvasPanel implements ActionListener {

	protected boolean autoScrollChatHistory = true;

	protected boolean initiator;

	protected boolean requestPen = false;

	protected boolean revokePen = false;

	protected boolean deniedPen = false;

	protected int timeout = -1;

	/**
	 * Straight text chat text field.
	 */
	protected JTextField chatTextField;

	/**
	 * Straight text chat history text pane.
	 */
	protected JTextPane chatHistoryTextPane;

	protected JScrollPane chatHistoryScrollPane;

	/**
	 * Jabber chat conversation thread identifier. This is non-null when the user
	 * is connected, has indicated a chat user that will receive messages sent,
	 * and at least one message has been sent to the user indicated. Otherwise,
	 * its null.
	 */
	protected String threadId = null;

	/**
	 * Jabber chat user who will receive messages. This is non-null when the user
	 * is connected and has indicated a chat user that will receive messages sent.
	 * <p>
	 * This field consists of the username, an ampersand, and the server
	 * (johndoe@jabber.org). The Jabbber resource is not included.
	 * (johndoe@jabber.org/Convey is incorrect.)
	 */
	protected String to = null;

	protected String subject = null;

	public JabberCanvasPanel(ConveyFrame conveyFrame, String to, String subject) {
		super(conveyFrame);
		super.setNetworked();

		this.to = to;

		if(subject != null)
			this.subject = subject;
		else
			this.subject = "";

		this.threadId = String.valueOf(System.currentTimeMillis());

		super.graphicPanel.setHasPen(false);
	}

	public JabberCanvasPanel(ConveyFrame conveyFrame, String to, String threadId, String subject) {
		super(conveyFrame);

		this.to = to;

		this.threadId = threadId;

		if(subject != null)
			this.subject = subject;
		else
			this.subject = "";

		super.graphicPanel.setHasPen(false);
	}

	public void initCenterComponent() {
		JPanel chatPanel = new JPanel(new BorderLayout(5, 5));

		this.chatHistoryTextPane = new JTextPane();
		this.chatHistoryTextPane.setEditable(false);
		this.chatTextField = new JTextField();
		this.chatTextField.addActionListener(this);

		this.chatHistoryScrollPane = new JScrollPane(this.chatHistoryTextPane,
				JScrollPane.VERTICAL_SCROLLBAR_ALWAYS,
				JScrollPane.HORIZONTAL_SCROLLBAR_NEVER);
		MouseListener ml = new MouseAdapter() {
			public void mouseReleased(MouseEvent e) {
				JScrollBar scrollBar = (JScrollBar)e.getSource();
				if(scrollBar.getValue() + scrollBar.getModel().getExtent() == scrollBar.getMaximum())
					autoScrollChatHistory = true;
				else
					autoScrollChatHistory = false;
			}
		};
		this.chatHistoryScrollPane.getVerticalScrollBar().addMouseListener(ml);

		chatPanel.add(this.chatHistoryScrollPane, BorderLayout.CENTER);

		Box chatMessageBox = new Box(BoxLayout.X_AXIS);
		JLabel messageLabel = new JLabel(" Message ");
		chatMessageBox.add(messageLabel);
		chatMessageBox.add(this.chatTextField);
		chatPanel.add(chatMessageBox, BorderLayout.SOUTH);

		JSplitPane splitPane = new JSplitPane(JSplitPane.VERTICAL_SPLIT, super.graphicPanel, chatPanel);
		splitPane.setDividerLocation(350);

		super.add(splitPane, BorderLayout.CENTER);
		super.setNetworked();
	}

	/**
	 * Sends straight text message if the user is connected and has indicated a
	 * chat receiver and has typed text into chatTextField.
	 */
	public void sendText() {
		JabberSession session = super.conveyFrame.getJabberSession();
		if(session != null && this.to != null) {
			if(this.chatTextField.getText().length() > 0) {
				JabberConveyMessage msg = new JabberConveyMessage();
				msg.setThreadID(this.threadId);
				if(this.subject != null)
					msg.setSubject(this.subject);
				msg.setTo(this.to);
				msg.setFrom(super.conveyFrame.getFrom());
				msg.setBody(this.chatTextField.getText());
				try {
					session.sendMessage(msg);
				} catch(SendMessageFailedException e) {
					e.printStackTrace();
				}

				this.appendMessage(msg, Color.blue);

				this.chatTextField.setText("");

				if(this.autoScrollChatHistory) {
					JScrollBar scrollBar = this.chatHistoryScrollPane.getVerticalScrollBar();
					scrollBar.setValue(scrollBar.getMaximum() - scrollBar.getModel().getExtent());
				}
			}
		}
	}

	/**
	 * Action event receiver for when the user presses &lt;Enter&gt; while the
	 * chat text field has focus.
	 */
	public void actionPerformed(ActionEvent e) {
		sendText();
	}

	public void appendMessage(JabberConveyMessage msg, Color nameColor) {
		StringTokenizer tokenizer = new StringTokenizer(msg.getFrom(), "@");
		StringBuffer name = new StringBuffer(" <");
		name.append(tokenizer.nextToken());
		name.append('>');
		DateFormat format = DateFormat.getDateTimeInstance(DateFormat.SHORT, DateFormat.MEDIUM);
		StringBuffer date = new StringBuffer(format.format(new Date()));
		date.insert(0, '[');
		date.append(']');
		StringBuffer message = new StringBuffer(" ");
		message.append(msg.getBody());
		message.append('\n');

		// Need package name to avoid conflict with org.jdom.Document
		javax.swing.text.Document doc = this.chatHistoryTextPane.getDocument();
		try {
			SimpleAttributeSet set = new SimpleAttributeSet();
			StyleConstants.setForeground(set, Color.lightGray);
			doc.insertString(doc.getLength(), date.toString(), set);
			StyleConstants.setForeground(set, nameColor);
			doc.insertString(doc.getLength(), name.toString(), set);
			StyleConstants.setForeground(set, Color.black);
			doc.insertString(doc.getLength(), message.toString(), set);
		} catch(Exception e) {
			e.printStackTrace();
		}
	}

	public String getThreadId() {
		return this.threadId;
	}

	public String getSubject() {
		return this.subject;
	}

	public String getTo() {
		return this.to;
	}

	public void handleMessage(JabberConveyMessage msg) {
		if(msg.isConvey())
			super.graphicPanel.handleConveyElement(msg.getConveyElement());
		else if(msg.isPen())
			this.handlePenElement(msg.getPenElement());
		else {
			this.appendMessage(msg, Color.red);

			if(this.autoScrollChatHistory) {
				JScrollBar scrollBar = this.chatHistoryScrollPane.getVerticalScrollBar();
				scrollBar.setValue(scrollBar.getMaximum() - scrollBar.getModel().getExtent());
			}
		}
	}

	public void sendGraphicEvent(GraphicEvent event) {
		JabberSession session = super.getConveyFrame().getJabberSession();
		if(session != null && this.to != null) {
			org.jdom.Element element = event.toElement();
			if(element != null) {
				org.jdom.Element conveyElement = new org.jdom.Element("convey", JabberConveyMessage.CONVEY_XMLNS);
				conveyElement.addContent(element);

				JabberConveyMessage msg = new JabberConveyMessage();
				msg.setThreadID(this.threadId);
				if(this.subject != null)
					msg.setSubject(this.subject);
				msg.setTo(this.to);
				msg.setFrom(super.getConveyFrame().getFrom());
				msg.setConveyElement(conveyElement);
				try {
					session.sendMessage(msg);
				} catch(SendMessageFailedException e) {
					e.printStackTrace();
				}
			}
		}
	}

	public void setInitiator(boolean value) {
		this.initiator = value;
	}

	public boolean isInitiator() {
		return this.initiator;
	}

	public void handlePenElement(org.jdom.Element penElement) {
		if(penElement.getChild("request") != null) {
			this.requestPen = true;
			if(this.getConveyFrame().getSelectedCanvasPanel() == this)
				doRequest();
		} else if(penElement.getChild("revoke") != null) {
			this.requestPen = false;
			this.revokePen = true;
			if(this.getConveyFrame().getSelectedCanvasPanel() == this)
				doRevoke();
		} else if(penElement.getChild("yield") != null) {
			this.requestPen = false;
			this.revokePen = false;
			super.graphicPanel.setHasPen(true);
		} else if(penElement.getChild("granted") != null) {
			this.requestPen = false;
			this.revokePen = false;
			super.graphicPanel.setHasPen(true);
		} else if(penElement.getChild("denied") != null) {
			this.deniedPen = true;
			if(this.getConveyFrame().getSelectedCanvasPanel() == this) {
				if(this.requestPen)
					doRequest();
			}
		}

		super.getConveyFrame().updateButtons();
	}

	public void doRequest() {
		if(super.graphicPanel.hasPen()) {
			int res = JOptionPane.showConfirmDialog(this, this.to + " is requesting the Pen. Do you wish to grant Pen request?", "Pen Request",
																							JOptionPane.YES_NO_OPTION, JOptionPane.QUESTION_MESSAGE,
																							ResourceLoader.loadImageIcon("org/convey/images/requestpen.png"));
			if(res == JOptionPane.YES_OPTION) {
				super.graphicPanel.flushGraphicEventQueue();
				super.graphicPanel.setHasPen(false);
				this.sendPenEvent(new PenEvent(PenEvent.GRANTED));
			} else
				this.sendPenEvent(new PenEvent(PenEvent.DENIED));
		} else if(this.deniedPen) {
			JOptionPane.showConfirmDialog(this, this.to + " denied your Pen request.", "Pen Request",
																		JOptionPane.OK_OPTION, JOptionPane.INFORMATION_MESSAGE,
																		ResourceLoader.loadImageIcon("org/convey/images/requestpen.png"));
			this.deniedPen = false;
		} else
			super.graphicPanel.setHasPen(true);

		this.requestPen = false;
	}

	public void doRevoke() {
		if(super.graphicPanel.hasPen()) {
			super.graphicPanel.flushGraphicEventQueue();
			super.graphicPanel.setHasPen(false);
			this.sendPenEvent(new PenEvent(PenEvent.YIELD));
			JOptionPane.showConfirmDialog(this, this.to + " is revoking the Pen.", "Pen Revoke",
																		JOptionPane.OK_OPTION, JOptionPane.INFORMATION_MESSAGE,
																		ResourceLoader.loadImageIcon("org/convey/images/revokepen.png"));
		}

		this.revokePen = false;
	}

	public boolean isRequest() {
		return this.requestPen;
	}

	public boolean isRevoke() {
		return this.revokePen;
	}

	public void sendPenEvent(PenEvent event) {
		JabberSession session = super.getConveyFrame().getJabberSession();
		if(session != null && this.to != null) {
			org.jdom.Element element = event.toElement();
			if(element != null) {
				org.jdom.Element conveyElement = new org.jdom.Element("pen", JabberConveyMessage.CONVEY_XMLNS);
				conveyElement.addContent(element);

				JabberConveyMessage msg = new JabberConveyMessage();
				msg.setThreadID(this.threadId);
				if(this.subject != null)
					msg.setSubject(this.subject);
				msg.setTo(this.to);
				msg.setFrom(super.getConveyFrame().getFrom());
				msg.setConveyElement(conveyElement);
				try {
					session.sendMessage(msg);
				} catch(SendMessageFailedException e) {
					e.printStackTrace();
				}
			}
		}
	}

	void penYield() {
		super.graphicPanel.flushGraphicEventQueue();

		super.graphicPanel.setHasPen(false);
		this.sendPenEvent(new PenEvent(PenEvent.YIELD));

		super.getConveyFrame().updateButtons();
	}

	void penRequest() {
		this.requestPen = true;
		this.sendPenEvent(new PenEvent(PenEvent.REQUEST));
	}

	void penRevoke() {
		this.requestPen = false;
		this.revokePen = true;
		this.sendPenEvent(new PenEvent(PenEvent.REVOKE));
	}

}