package org.convey;

import java.util.*;

import java.awt.*;

import java.awt.event.*;

import java.awt.geom.*;

import org.jdom.*;

/**
 * @author Andy Ames
 * @version 0.2
 */
public class PolygonGraphic extends ShapeGraphic {

	protected Vector points = new Vector();

	protected Vector origPoints = new Vector();

	private transient int pointMoving = -1;

	public PolygonGraphic() {
		this.name = "polygon1";
	}

	public PolygonGraphic(GraphicPanel graphicPanel) {
		super(graphicPanel);
		this.name = "polygon1";
	}

	public PolygonGraphic(GraphicPanel graphicPanel, Point p) {
		super(graphicPanel);
		points.add(new Point2D.Double((double)p.x, (double)p.y));
		points.add(new Point2D.Double((double)p.x, (double)p.y));
		updateCache();
		super.repaint();
		this.name = "polygon1";
	}

	public Element toElement() {
		Element element = super.toElement();
		element.setName("polygon");

		StringBuffer buffer = new StringBuffer();
		Iterator iter = points.iterator();
		while(iter.hasNext()) {
			Point2D p = (Point2D)iter.next();
			buffer.append(p.getX());
			buffer.append(",");
			buffer.append(p.getY());
			buffer.append(" ");
		}
		if(buffer.length() > 0)
			buffer.deleteCharAt(buffer.length() - 1);

		element.setAttribute("points", buffer.toString());
		return element;
	}

	public void fromElement(Element element) {
		super.fromElement(element);

		StringTokenizer tokenizer = new StringTokenizer(element.getAttributeValue("points"), " ,");
		this.points.clear();
		while(tokenizer.hasMoreTokens()) {
			double x = Double.parseDouble(tokenizer.nextToken());
			double y = Double.parseDouble(tokenizer.nextToken());
			this.points.add(new Point2D.Double(x, y));
		}
		this.updateCache();
		this.repaint();
	}

	public void paintSelection(Graphics2D g2d) {
		switch(super.graphicPanel.getState()) {
			case GraphicPanel.SCALE:
				super.paintSelection(g2d);
				break;
			default:
				g2d.setColor(Color.black);
				g2d.setStroke(new BasicStroke());

				Iterator iter = this.points.iterator();
				while(iter.hasNext()) {
					Point2D p = (Point2D)iter.next();
					g2d.fillRect((int)p.getX() - 2, (int)p.getY() - 2, 5, 5);
				}
				break;
		}
	}

	public boolean hitVertex(Point p) {
		if(super.getState() == Graphic.CREATING)
			return false;
		switch(this.graphicPanel.getState()) {
			case GraphicPanel.SCALE:
				return super.hitVertex(p);
			default:
				Iterator iter = this.points.iterator();
				while(iter.hasNext()) {
					Point2D v = (Point2D)iter.next();
					double x = v.getX();
					double y = v.getY();
					if(p.x >= x - 3 && p.x <= x + 3 && p.y >= y - 3 && p.y <= y + 3)
						return true;
				}

				break;
		}

		return false;
	}

	public void fitToRect(Rectangle r) {
		if(this.origBounds.width == 0 || this.origBounds.height == 0)
			return;
		double tx1 = (double)-super.origBounds.x;
		double ty1 = (double)-super.origBounds.y;
		double sx = (double)r.width / (double)super.origBounds.width;
		double sy = (double)r.height / (double)super.origBounds.height;
		double tx2 = (double)r.x;
		double ty2 = (double)r.y;

		Iterator origIter = this.origPoints.iterator();
		Iterator iter = this.points.iterator();
		while(iter.hasNext()) {
			Point2D origP = (Point2D)origIter.next();
			Point2D p = (Point2D)iter.next();
			p.setLocation(((origP.getX() + tx1) * sx + tx2),
										((origP.getY() + ty1) * sy + ty2));
		}

		super.fitToRect(r);
	}

	/**
	 * Invoked when a mouse button has been pressed on a component.
	 */
	public void mousePressed(MouseEvent e) {
		if(super.getState() == Graphic.CREATING)
			return;
		Point p = e.getPoint();
		switch(super.graphicPanel.getState()) {
			case GraphicPanel.SCALE:
				super.mousePressed(e);
				break;
			default:
				Iterator iter = this.points.iterator();
				int i = 0;
				while(iter.hasNext()) {
					Point2D pt = (Point2D)iter.next();
					double x = pt.getX();
					double y = pt.getY();
					if(p.x >= x - 3 && p.x <= x + 3 &&
						 p.y >= y - 3 && p.y <= y + 3) {
						pointMoving = i;
						break;
					}
					i++;
				}

				break;
		}

		pressPoint = lastDragPoint = p;
	}

	public void setOrig() {
		super.setOrig();

		this.origPoints.clear();
		Iterator iter = this.points.iterator();
		while(iter.hasNext())
			this.origPoints.add(((Point2D)iter.next()).clone());
	}

	/**
	 * Invoked when the mouse has been clicked on a component.
	 */
	public void mouseClicked(MouseEvent e) {
		if(super.getState() == CREATING) {
			if((e.isControlDown()) || (e.isAltDown()) || (e.isMetaDown()) ||
				 (e.isShiftDown())) { /* stopping key detected */
				super.setState(DORMANT);
			} else {
				Point p = e.getPoint();
				points.add(new Point2D.Double((double)p.x, (double)p.y));
			}

			updateCache();
			super.repaint();
		}
	}

	/**
	 * Invoked when the mouse button has been moved on a component
	 * (with no buttons or mouse buttons pressed)
	 */
	public void mouseMoved(MouseEvent e) {
		if(super.getState() == CREATING) {
			Point2D pt = (Point2D)points.elementAt(points.size() - 1);
			Point p = e.getPoint();
			pt.setLocation((double)p.x, (double)p.y);

			updateCache();
			super.repaint();
		}
	}

	/**
	* Invoked when a mouse button is pressed on a component and then
	* dragged.  Mouse drag events will continue to be delivered to
	* the component where the first originated until the mouse button is
	* released (regardless of whether the mouse position is within the
	* bounds of the component).
	*/
	public void mouseDragged(MouseEvent e) {
		if(pointMoving != -1) {
			Point p = e.getPoint();
			points.remove(pointMoving);
			points.insertElementAt(p, pointMoving);
			this.updateCache();
			super.repaint();
			lastDragPoint = p;
		} else
			super.mouseDragged(e);
	}

	/**
	 * Invoked when a mouse button has been released on a component.
	 */
	public void mouseReleased(MouseEvent e) {
		if(pointMoving != -1) {
			Point p = e.getPoint();
			points.remove(pointMoving);
			points.insertElementAt(new Point2D.Double(p.x, p.y), pointMoving);

			Point2D origPoint = (Point2D)this.origPoints.elementAt(pointMoving);
			PolygonGraphicEvent event = new PolygonGraphicEvent(this, pointMoving, origPoint, new Point2D.Double(p.x, p.y));
			super.graphicPanel.pushGraphicEvent(event);

			pointMoving = -1;
		} else
			super.mouseReleased(e);
	}

	/**
	 *
	 */
	public void updateCache() {
		Point2D p = (Point2D)this.points.get(0);

		super.bounds = new Rectangle((int)p.getX(), (int)p.getY(), 0, 0);
		GeneralPath path = new GeneralPath();
		path.moveTo((float)p.getX(), (float)p.getY());

		for(int i = 1; i < this.points.size(); i++) {
			p = (Point2D)this.points.get(i);
			super.bounds.add((int)p.getX(), (int)p.getY());
			path.lineTo((float)p.getX(), (float)p.getY());
		}
		path.closePath();
		super.shape = path;
		Area area = new Area(new BasicStroke(6).createStrokedShape(super.shape));
		if(super.fillColor != null)
			area.add(new Area(super.shape));
		super.pathHotSpot = area;
		super.currDirty = (Rectangle)super.bounds.clone();
		super.currDirty.grow(5, 5);
	}

	public String getStatus() {
		return "";
	}

	public void translate(int x, int y) {
		super.translate(x, y);
		Iterator iter = points.iterator();
		while(iter.hasNext()) {
			Point2D p = (Point2D)iter.next();
			p.setLocation(p.getX() + (double)x, p.getY() + (double)y);
		}
		super.repaint();
	}

	public Vector getPoints() {
		return (Vector)this.points.clone();
	}

	public void setPoints(Vector points) {
		this.points = points;
		this.updateCache();
		super.repaint();
	}

	public void undoGraphicEvent(GraphicEvent event) {
		super.undoGraphicEvent(event);

		if(event instanceof PolygonGraphicEvent) {
			PolygonGraphicEvent polygonGraphicEvent = (PolygonGraphicEvent)event;

			Point2D point = (Point2D)this.points.elementAt(polygonGraphicEvent.getIndex());
			Point2D origPoint = polygonGraphicEvent.getOrigPoint();
			point.setLocation(origPoint);

			this.updateCache();
			super.repaint();
		}
	}

	public void redoGraphicEvent(GraphicEvent event) {
		super.redoGraphicEvent(event);

		if(event instanceof PolygonGraphicEvent) {
			PolygonGraphicEvent polygonGraphicEvent = (PolygonGraphicEvent)event;

			Point2D point = (Point2D)this.points.elementAt(polygonGraphicEvent.getIndex());
			Point2D newPoint = polygonGraphicEvent.getNewPoint();
			point.setLocation(newPoint);

			this.updateCache();
			super.repaint();
		}
	}

}