package org.convey;

import java.util.*;

import java.awt.*;

import java.awt.event.*;

import java.awt.geom.*;

import org.jdom.*;

/**
 * @author Andy Ames
 * @version 0.2
 */
public class QuadCurveGraphic extends ShapeGraphic {

	protected QuadCurve2D curve;

	protected QuadCurve2D origCurve;

	protected transient int clickCount;

	private transient int pointMoving = -1;

	private transient boolean repaintedControls = true;

	public QuadCurveGraphic() {
		this.name = "quadCurve1";
	}

	public QuadCurveGraphic(GraphicPanel graphicPanel) {
		super(graphicPanel);
		this.name = "quadCurve1";
	}

	public QuadCurveGraphic(GraphicPanel graphicPanel, Point p) {
		super(graphicPanel);
		curve = new QuadCurve2D.Double((double)p.x, (double)p.y,
											(double)p.x, (double)p.y,
											(double)p.x, (double)p.y);
		clickCount = 1;
		updateCache();
		super.repaint();
		this.name = "quadCurve1";
	}

	public Element toElement() {
		Element element = super.toElement();

		element.setName("path");

		StringBuffer buffer = new StringBuffer();
		buffer.append("M");
		buffer.append((int)curve.getX1());
		buffer.append(",");
		buffer.append((int)curve.getY1());
		buffer.append(" ");
		buffer.append("Q");
		buffer.append((int)curve.getCtrlX());
		buffer.append(",");
		buffer.append((int)curve.getCtrlY());
		buffer.append(" ");
		buffer.append((int)curve.getX2());
		buffer.append(",");
		buffer.append((int)curve.getY2());

		element.setAttribute("d", buffer.toString());

		return element;
	}

	public void fromElement(Element element) {
		super.fromElement(element);

		StringTokenizer tokenizer = new StringTokenizer(element.getAttributeValue("d"), "M, Q");

		int X1 = Integer.parseInt(tokenizer.nextToken());
		int Y1 = Integer.parseInt(tokenizer.nextToken());
		int ctrlX = Integer.parseInt(tokenizer.nextToken());
		int ctrlY = Integer.parseInt(tokenizer.nextToken());
		int X2 = Integer.parseInt(tokenizer.nextToken());
		int Y2 = Integer.parseInt(tokenizer.nextToken());

		this.curve = new QuadCurve2D.Float(X1, Y1, ctrlX, ctrlY, X2, Y2);
		this.updateCache();
		this.repaint();
	}

	public void fitToRect(Rectangle r) {
		if(this.origBounds.width == 0 || this.origBounds.height == 0)
			return;
		double tx1 = (double)-super.origBounds.x;
		double ty1 = (double)-super.origBounds.y;
		double sx = (double)r.width / (double)super.origBounds.width;
		double sy = (double)r.height / (double)super.origBounds.height;
		double tx2 = (double)r.x;
		double ty2 = (double)r.y;

		double x1 = (this.origCurve.getX1() + tx1) * sx + tx2;
		double y1 = (this.origCurve.getY1() + ty1) * sy + ty2;
		double ctrlX = (this.origCurve.getCtrlX() + tx1) * sx + tx2;
		double ctrlY = (this.origCurve.getCtrlY() + ty1) * sy + ty2;
		double x2 = (this.origCurve.getX2() + tx1) * sx + tx2;
		double y2 = (this.origCurve.getY2() + ty1) * sy + ty2;

		this.curve.setCurve(x1, y1, ctrlX, ctrlY, x2, y2);

		super.fitToRect(r);
	}

	public void paintSelection(Graphics2D g2d) {
		switch(super.graphicPanel.getState()) {
			case GraphicPanel.SCALE:
				super.paintSelection(g2d);
				break;
			default:
				g2d.setColor(Color.gray);
				float[] dash = { 6.0f, 6.0f };
				g2d.setStroke(new BasicStroke(1.0f, BasicStroke.CAP_ROUND, BasicStroke.JOIN_ROUND, 0.0f, dash, 0.0f));

				g2d.drawLine((int)curve.getX1(), (int)curve.getY1(), (int)curve.getCtrlX(), (int)curve.getCtrlY());
				g2d.drawLine((int)curve.getX2(), (int)curve.getY2(), (int)curve.getCtrlX(), (int)curve.getCtrlY());

				g2d.setColor(Color.black);
				g2d.setStroke(new BasicStroke());
				g2d.fillRect((int)curve.getCtrlX() - 2, (int)curve.getCtrlY() - 2, 5, 5);
				g2d.fillRect((int)curve.getX1() - 2, (int)curve.getY1() - 2, 5, 5);
				g2d.fillRect((int)curve.getX2() - 2, (int)curve.getY2() - 2, 5, 5);
				break;
		}
	}

	public boolean hitVertex(Point p) {
		if(super.getState() == Graphic.CREATING)
			return false;
		switch(this.graphicPanel.getState()) {
			case GraphicPanel.SCALE:
				return super.hitVertex(p);
			default:
				int x1 = (int)this.curve.getX1();
				int y1 = (int)this.curve.getY1();
				int x2 = (int)this.curve.getX2();
				int y2 = (int)this.curve.getY2();
				int ctrlX = (int)this.curve.getCtrlX();
				int ctrlY = (int)this.curve.getCtrlY();
				if(p.x >= x1 - 3 && p.x <= x1 + 3 && p.y >= y1 - 3 && p.y <= y1 + 3)
					return true;
				else if(p.x >= x2 - 3 && p.x <= x2 + 3 && p.y >= y2 - 3 && p.y <= y2 + 3)
					return true;
				else if(p.x >= ctrlX - 3 && p.x <= ctrlX + 3 && p.y >= ctrlY - 3 && p.y <= ctrlY + 3)
					return true;

				break;
		}

		return false;
	}

	/**
	 * Invoked when a mouse button has been pressed on a component.
	 */
	public void mousePressed(MouseEvent e) {
		if(super.getState() == Graphic.CREATING)
			return;
		Point p = e.getPoint();
		switch(super.graphicPanel.getState()) {
			case GraphicPanel.SCALE:
				super.mousePressed(e);
				break;
			default:
				if(p.x >= curve.getCtrlX() - 3 && p.x <= curve.getCtrlX() + 3 &&
					 p.y >= curve.getCtrlY() - 3 && p.y <= curve.getCtrlY() + 3)
					pointMoving = 1;
				else if(p.x >= curve.getX1() - 3 && p.x <= curve.getX1() + 3 &&
								p.y >= curve.getY1() - 3 && p.y <= curve.getY1() + 3)
					pointMoving = 0;
				else if(p.x >= curve.getX2() - 3 && p.x <= curve.getX2() + 3 &&
								p.y >= curve.getY2() - 3 && p.y <= curve.getY2() + 3)
					pointMoving = 2;

				break;
		}

		pressPoint = lastDragPoint = p;
	}

	public void setOrig() {
		super.setOrig();
		this.origCurve = (QuadCurve2D)this.curve.clone();
	}

	/**
	 * Invoked when the mouse has been clicked on a component.
	 */
	public void mouseClicked(MouseEvent e) {
		if(super.getState() == CREATING) {
			clickCount++;
			Point p = e.getPoint();
			if(clickCount == 2)
				curve.setCurve(curve.getX1(), curve.getY1(), (double)p.x, (double)p.y,
											 (double)p.x, (double)p.y);
			else if(clickCount == 3) {
				curve.setCurve(curve.getX1(), curve.getY1(),
											 (double)p.x, (double)p.y,
											 curve.getX2(), curve.getY2());
				super.setState(DORMANT);
			}

			updateCache();
			super.repaint();
		}
	}

	/**
	 * Invoked when the mouse button has been moved on a component
	 * (with no buttons or mouse buttons pressed)
	 */
	public void mouseMoved(MouseEvent e) {
		if(super.getState() == CREATING) {
			Point p = e.getPoint();
			if(clickCount == 1)
				curve.setCurve(curve.getX1(), curve.getY1(), (double)p.x, (double)p.y,
											 (double)p.x, (double)p.y);
			else if(clickCount == 2)
				curve.setCurve(curve.getX1(), curve.getY1(),
											 (double)p.x, (double)p.y,
											 curve.getX2(), curve.getY2());

			updateCache();
			super.repaint();
		}
	}

	/**
	* Invoked when a mouse button is pressed on a component and then
	* dragged.  Mouse drag events will continue to be delivered to
	* the component where the first originated until the mouse button is
	* released (regardless of whether the mouse position is within the
	* bounds of the component).
	*/
	public void mouseDragged(MouseEvent e) {
		if(pointMoving != -1) {
			Point p = e.getPoint();
			if(pointMoving == 1)
				curve.setCurve(curve.getX1(), curve.getY1(),
											 curve.getCtrlX() + p.x - lastDragPoint.x,
											 curve.getCtrlY() + p.y - lastDragPoint.y,
											 curve.getX2(), curve.getY2());
			else if(pointMoving == 0)
				curve.setCurve(curve.getX1() + p.x - lastDragPoint.x,
											 curve.getY1() + p.y - lastDragPoint.y,
											 curve.getCtrlX(), curve.getCtrlY(),
											 curve.getX2(), curve.getY2());
			else if(pointMoving == 2)
				curve.setCurve(curve.getX1(), curve.getY1(),
											 curve.getCtrlX(), curve.getCtrlY(),
											 curve.getX2() + p.x - lastDragPoint.x,
											 curve.getY2() + p.y - lastDragPoint.y);
			this.updateCache();
			super.repaint();
			lastDragPoint = p;
		} else
			super.mouseDragged(e);
	}

	/**
	 * Invoked when a mouse button has been released on a component.
	 */
	public void mouseReleased(MouseEvent e) {
		if(pointMoving != -1) {
			Point p = e.getPoint();

			QuadCurveGraphicEvent event = new QuadCurveGraphicEvent(this, this.origCurve, this.curve);
			super.graphicPanel.pushGraphicEvent(event);

			pointMoving = -1;
		} else
			super.mouseReleased(e);
	}

	/**
	 * @todo Implement an exact bounding rectangle for the curve
	 */
	public void updateCache() {
		super.shape = (Shape)this.curve.clone();
		Area area = new Area(new BasicStroke(1).createStrokedShape(super.shape));
		super.bounds = area.getBounds();
		area = new Area(new BasicStroke(6).createStrokedShape(super.shape));
		if(super.fillColor != null)
			area.add(new Area(super.shape));
		super.pathHotSpot = area;

		// When the curve is in a group, its control points will never be included
		// in the hotspot or bounds because a Graphic is never directly selected
		// while in a group.
		if(super.getState() == SELECTED) {
			switch(super.graphicPanel.getState()) {
				case GraphicPanel.SCALE:
					break;
				default:
					area.add(new Area(new Rectangle((int)curve.getCtrlX() - 3, (int)curve.getCtrlY() - 3, 7, 7)));
					area.add(new Area(new Rectangle((int)curve.getX1() - 3, (int)curve.getY1() - 3, 7, 7)));
					area.add(new Area(new Rectangle((int)curve.getX2() - 3, (int)curve.getY2() - 3, 7, 7)));
					repaintedControls = false;
					break;
			}
		} else if(!repaintedControls) {
			area.add(new Area(new Rectangle((int)curve.getCtrlX() - 3, (int)curve.getCtrlY() - 3, 7, 7)));
			area.add(new Area(new Rectangle((int)curve.getX1() - 3, (int)curve.getY1() - 3, 7, 7)));
			area.add(new Area(new Rectangle((int)curve.getX2() - 3, (int)curve.getY2() - 3, 7, 7)));
			repaintedControls = true;
		}

		super.currDirty = area.getBounds();
		super.currDirty.grow(5, 5);
	}

	public String getStatus() {
		return "";
	}

	public void translate(int x, int y) {
		super.translate(x, y);
		curve.setCurve(curve.getX1() + (double)x, curve.getY1() + (double)y,
									 curve.getCtrlX() + (double)x, curve.getCtrlY() + (double)y,
									 curve.getX2() + (double)x, curve.getY2() + (double)y);
		super.repaint();
	}

	public void setCurve(QuadCurve2D curve) {
		this.curve = (QuadCurve2D)curve.clone();
		this.updateCache();
		super.repaint();
	}

	public void undoGraphicEvent(GraphicEvent event) {
		super.undoGraphicEvent(event);

		if(event instanceof QuadCurveGraphicEvent) {
			QuadCurveGraphicEvent cubicCurveGraphicEvent = (QuadCurveGraphicEvent)event;

			this.setCurve(cubicCurveGraphicEvent.getOrigCurve());
		}
	}

	public void redoGraphicEvent(GraphicEvent event) {
		super.redoGraphicEvent(event);

		if(event instanceof QuadCurveGraphicEvent) {
			QuadCurveGraphicEvent cubicCurveGraphicEvent = (QuadCurveGraphicEvent)event;

			this.setCurve(cubicCurveGraphicEvent.getNewCurve());
		}
	}

}