package org.convey;

import java.awt.*;

import java.awt.geom.*;

import java.awt.event.*;

import org.jdom.*;

/**
 * @author Andy Ames
 * @version 0.2
 */
public class QuadGraphic extends ShapeGraphic {

	protected Rectangle quad;

	protected transient Rectangle origQuad;

	protected transient int vertexMoving = -1;

	public QuadGraphic() {
		this.name = "quad1";
	}

	public QuadGraphic(GraphicPanel graphicPanel) {
		super(graphicPanel);
		this.name = "quad1";
	}

	public QuadGraphic(GraphicPanel graphicPanel, Point p) {
		super(graphicPanel);
		this.quad = new Rectangle(p.x, p.y, 0, 0);
		updateCache();
		super.repaint();
		this.name = "quad1";
	}

	public void paintSelection(Graphics2D g2d) {
		switch(super.graphicPanel.getState()) {
			case GraphicPanel.SCALE:
				super.paintSelection(g2d);
				break;
			default:
				g2d.setColor(Color.black);
				g2d.setStroke(new BasicStroke(1));
				g2d.fillRect(this.quad.x - 2, this.quad.y - 2, 5, 5);
				g2d.fillRect(this.quad.x + this.quad.width - 2, this.quad.y - 2, 5, 5);
				g2d.fillRect(this.quad.x + this.quad.width - 2, this.quad.y + this.quad.height - 2, 5, 5);
				g2d.fillRect(this.quad.x - 2 , this.quad.y + this.quad.height - 2, 5, 5);
				break;
		}
	}

	public Element toElement() {
		Element element = super.toElement();

		element.setName("rect");

		element.setAttribute("x", String.valueOf(quad.x));
		element.setAttribute("y", String.valueOf(quad.y));
		element.setAttribute("width", String.valueOf(quad.width));
		element.setAttribute("height", String.valueOf(quad.height));

		return element;
	}

	public void fromElement(Element element) {
		super.fromElement(element);

		int X = Integer.parseInt(element.getAttributeValue("x"));
		int Y = Integer.parseInt(element.getAttributeValue("y"));
		int width = Integer.parseInt(element.getAttributeValue("width"));
		int height = Integer.parseInt(element.getAttributeValue("height"));

		this.quad = new Rectangle(X, Y, width, height);
		this.updateCache();
		this.repaint();
	}

	public boolean hitVertex(Point p) {
		if(super.getState() == Graphic.CREATING)
			return false;
		switch(this.graphicPanel.getState()) {
			case GraphicPanel.SCALE:
				return super.hitVertex(p);
			default:
				int xcoin = -1; // 0 is left, 1 is right, 2 is center
				int ycoin = -1; // 0 is top, 1 is bottom, 2 is middle

				if(p.x >= this.quad.x - 3 && p.x <= this.quad.x + 3) {
					// left edge
					xcoin = 0;
				} else if(p.x >= this.quad.x + this.quad.width - 3 && p.x <= this.quad.x + this.quad.width + 3) {
					// right edge
					xcoin = 1;
				}  else if(p.x >= this.quad.x + this.quad.width / 2 - 3 && p.x <= this.quad.x + this.quad.width / 2 + 3) {
					// center
					xcoin = 2;
				}

				if(xcoin == -1)
					return false;

				if(p.y >= this.quad.y - 3 && p.y <= this.quad.y + 3) {
					// top edge
					ycoin = 0;
				} else if(p.y >= this.quad.y + this.quad.height - 3 && p.y <= this.quad.y + this.quad.height + 3) {
					// bottom edge
					ycoin = 1;
				}  else if(p.y >= this.quad.y + this.quad.height / 2 - 3 && p.y <= this.quad.y + this.quad.height / 2 + 3) {
					// middle
					ycoin = 2;
				}

				if((xcoin == 0 || xcoin == 1) && (ycoin == 0 || ycoin == 1))
					return true;

				break;
		}

		return false;
	}

	public void mousePressed(MouseEvent e) {
		if(super.getState() == Graphic.CREATING)
			return;
		switch(super.graphicPanel.getState()) {
			case GraphicPanel.SCALE:
				super.mousePressed(e);
				break;
			default:
				Point p = e.getPoint();
				int xcoin = -1; // -1 is not coincident, 0 is left, 1 is right
				int ycoin = -1; // -1 is not coincident, 0 is top, 1 is bottom

				if(p.x >= this.bounds.x - 3 && p.x <= this.bounds.x + 3) {
					// left edge
					xcoin = 0;
				} else if(p.x >= this.bounds.x + this.bounds.width - 3 && p.x <= this.bounds.x + this.bounds.width + 3) {
					// right edge
					xcoin = 1;
				}

				if(p.y >= this.bounds.y - 3 && p.y <= this.bounds.y + 3) {
					// top edge
					ycoin = 0;
				} else if(p.y >= this.bounds.y + this.bounds.height - 3 && p.y <= this.bounds.y + this.bounds.height + 3) {
					// bottom edge
					ycoin = 1;
				}

				scaleAnchor = -1;
				if(xcoin == 0) {
					if(ycoin == 0)
						super.scaleAnchor = BOTTOM_RIGHT;
					else if(ycoin == 1)
						super.scaleAnchor = TOP_RIGHT;
				} else if(xcoin == 1) {
					if(ycoin == 0)
						super.scaleAnchor = BOTTOM_LEFT;
					else if(ycoin == 1)
						super.scaleAnchor = TOP_LEFT;
				}
				break;
		}

		lastDragPoint = pressPoint = e.getPoint();
	}

	public void setOrig() {
		super.setOrig();
		this.origQuad = (Rectangle)this.quad.clone();
	}

	public void fitToRect(Rectangle r) {
		this.quad = (Rectangle)r.clone();

		super.fitToRect(r);
	}

	/**
	 * Invoked when a mouse button has been released on a component.
	 */
	public void mouseReleased(MouseEvent e) {
		if(super.getState() == CREATING) {
			Point p = e.getPoint();
			quad.width = p.x - quad.x;
			quad.height = p.y - quad.y;
			super.setState(DORMANT);
		} else if(this.scaleAnchor != -1) {
			if(super.lastFitToRect != null) {
				this.graphicPanel.pushGraphicEvent(new FitToRectGraphicEvent(this, this.origBounds, this.lastFitToRect));
				this.scaleAnchor = -1;
				this.lastFitToRect = null;
			}
		} else
			super.mouseReleased(e);
	}

	/**
	* Invoked when a mouse button is pressed on a component and then
	* dragged.  Mouse drag events will continue to be delivered to
	* the component where the first originated until the mouse button is
	* released (regardless of whether the mouse position is within the
	* bounds of the component).
	*/
	public void mouseDragged(MouseEvent e) {
		if(super.getState() == CREATING) {
			Point p = e.getPoint();
			quad.width = p.x - quad.x;
			quad.height = p.y - quad.y;

			updateCache();
			super.repaint();
		} else {
			Point p = e.getPoint();
			switch(super.graphicPanel.getState()) {
				case GraphicPanel.SCALE:
					super.mouseDragged(e);
					break;
				default:
					if(scaleAnchor != -1) {
						super.lastFitToRect = (Rectangle)super.origBounds.clone();
						switch(scaleAnchor) {
							case TOP_LEFT:
								this.lastFitToRect.x = this.origBounds.x;
								this.lastFitToRect.width = p.x - this.origBounds.x;

								this.lastFitToRect.y = this.origBounds.y;
								this.lastFitToRect.height = p.y - this.origBounds.y;
								break;
							case TOP_RIGHT:
								this.lastFitToRect.x = p.x;
								this.lastFitToRect.width = this.origBounds.x + this.origBounds.width - p.x;

								this.lastFitToRect.y = this.origBounds.y;
								this.lastFitToRect.height = p.y - this.origBounds.y;
								break;
							case BOTTOM_RIGHT:
								this.lastFitToRect.x = p.x;
								this.lastFitToRect.width = this.origBounds.x + this.origBounds.width - p.x;

								this.lastFitToRect.y = p.y;
								this.lastFitToRect.height = this.origBounds.y + this.origBounds.height - p.y;
								break;
							case BOTTOM_LEFT:
								this.lastFitToRect.x = this.origBounds.x;
								this.lastFitToRect.width = p.x - this.origBounds.x;

								this.lastFitToRect.y = p.y;
								this.lastFitToRect.height = this.origBounds.y + this.origBounds.height - p.y;
								break;
						}
						this.fitToRect(this.lastFitToRect);
						this.updateCache();
						super.repaint();
					} else {
						this.translate(p.x - lastDragPoint.x, p.y - lastDragPoint.y);
						lastDragPoint = p;
					}
					break;
			}
		}
	}

	public void updateCache() {
		int x = Math.min(quad.x, quad.x + quad.width);
		int y = Math.min(quad.y, quad.y + quad.height);
		int w = Math.abs(quad.width);
		int h = Math.abs(quad.height);
		super.shape = new Rectangle(x, y, w, h);
		Area area = new Area(new BasicStroke(6).createStrokedShape(super.shape));
		if(super.fillColor != null)
			area.add(new Area(super.shape));
		super.pathHotSpot = area;
		super.bounds = new Rectangle(x, y, w, h);
		super.currDirty = new Rectangle(x, y, w, h);
		super.currDirty.grow(5, 5);
	}

	public String getStatus() {
		return "RECTANGLE: (" + quad.x + "," + quad.y + "," + quad.width + "," + quad.height + ")";
	}

	public void translate(int x, int y) {
		super.translate(x, y);
		quad.translate(x, y);
		super.repaint();
	}

	public void setQuad(Rectangle quad) {
		this.quad = (Rectangle)quad.clone();
		this.updateCache();
		super.repaint();
	}

	public void undoGraphicEvent(GraphicEvent event) {
		super.undoGraphicEvent(event);

		if(event instanceof QuadGraphicEvent) {
			QuadGraphicEvent ellipseGraphicEvent = (QuadGraphicEvent)event;

			this.setQuad(ellipseGraphicEvent.getOrigQuad());
		}
	}

	public void redoGraphicEvent(GraphicEvent event) {
		super.redoGraphicEvent(event);

		if(event instanceof QuadGraphicEvent) {
			QuadGraphicEvent ellipseGraphicEvent = (QuadGraphicEvent)event;

			this.setQuad(ellipseGraphicEvent.getNewQuad());
		}
	}

}