package org.convey;

import javax.swing.*;
import javax.swing.border.*;

import java.awt.*;
import java.awt.event.*;

import java.util.*;

/**
 * @author Andy Ames
 */
public class ShapeGraphicPropertyPanel extends PropertyPanel {

	protected JDialog fillColorChooserDialog;

	protected JColorChooser fillColorChooser;

	protected JDialog strokeColorChooserDialog;

	protected JColorChooser strokeColorChooser;

	protected JPanel fillColorPanel;

	protected JCheckBox filledCheckBox;

	protected JPanel strokeColorPanel;

	protected JCheckBox strokedCheckBox;

	protected boolean noneSelected;

	public ShapeGraphicPropertyPanel() {
		super("Shape");

		super.setSize(150, 155);
		super.setPreferredSize(new Dimension(150, 155));
		super.setMaximumSize(new Dimension(150, 155));
		super.setMinimumSize(new Dimension(150, 155));

		((FlowLayout)super.getLayout()).setAlignment(FlowLayout.LEFT);

		Box box = new Box(BoxLayout.Y_AXIS);
		super.add(box);

		Box fillCheckBoxBox = new Box(BoxLayout.X_AXIS);

		this.filledCheckBox = new JCheckBox("Filled");
		ActionListener al = new ActionListener() {
			public void actionPerformed(ActionEvent e) {
				setFillColor();
			}
		};
		this.filledCheckBox.addActionListener(al);

		MouseListener ml = new MouseAdapter() {
			public void mouseClicked(MouseEvent e) {
				if(selectedGraphics != null) {
					filledCheckBox.setEnabled(true);
					setFillColor();
				}
			}
		};
		this.filledCheckBox.addMouseListener(ml);

		fillCheckBoxBox.add(this.filledCheckBox);
		fillCheckBoxBox.add(Box.createHorizontalGlue());
		box.add(fillCheckBoxBox);

		this.fillColorPanel = new JPanel();
		this.fillColorPanel.setBackground(Color.white);
		this.fillColorPanel.setSize(32, 32);
		this.fillColorPanel.setMaximumSize(new Dimension(32, 32));
		this.fillColorPanel.setMinimumSize(new Dimension(32, 32));
		this.fillColorPanel.setPreferredSize(new Dimension(32, 32));

		final ShapeGraphicPropertyPanel thisPanel = this;
		final ActionListener finalAl1 = new ActionListener() {
			public void actionPerformed(ActionEvent e) {
				fillColorPanel.setBackground(fillColorChooser.getColor());
				if(filledCheckBox.isSelected())
					setFillColor();
			}
		};
		ml = new MouseAdapter() {
			public void mouseClicked(MouseEvent e) {
				if(noneSelected || selectedGraphics != null) {
					if(fillColorChooserDialog == null) {
						fillColorChooser = new JColorChooser(fillColorPanel.getBackground());
						fillColorChooserDialog = JColorChooser.createDialog(thisPanel,
							"Fill Color", true, fillColorChooser, finalAl1, null);
					}
					else
						fillColorChooser.setColor(fillColorPanel.getBackground());
					fillColorChooserDialog.show();
				}
			}
		};
		this.fillColorPanel.addMouseListener(ml);

		box.add(this.fillColorPanel);

		Box strokeCheckBoxBox = new Box(BoxLayout.X_AXIS);

		this.strokedCheckBox = new JCheckBox("Stroked");
		al = new ActionListener() {
			public void actionPerformed(ActionEvent e) {
				setStrokeColor();
			}
		};
		this.strokedCheckBox.setSelected(true);
		this.strokedCheckBox.addActionListener(al);

		ml = new MouseAdapter() {
			public void mouseClicked(MouseEvent e) {
				if(selectedGraphics != null) {
					strokedCheckBox.setEnabled(true);
					setStrokeColor();
				}
			}
		};
		this.strokedCheckBox.addMouseListener(ml);

		strokeCheckBoxBox.add(this.strokedCheckBox);
		strokeCheckBoxBox.add(Box.createHorizontalGlue());
		box.add(strokeCheckBoxBox);

		this.strokeColorPanel = new JPanel();
		this.strokeColorPanel.setBackground(Color.black);
		this.strokeColorPanel.setSize(32, 32);
		this.strokeColorPanel.setMaximumSize(new Dimension(32, 32));
		this.strokeColorPanel.setMinimumSize(new Dimension(32, 32));
		this.strokeColorPanel.setPreferredSize(new Dimension(32, 32));

		final ActionListener finalAl2 = new ActionListener() {
			public void actionPerformed(ActionEvent e) {
				strokeColorPanel.setBackground(strokeColorChooser.getColor());
				if(strokedCheckBox.isSelected())
					setStrokeColor();
			}
		};
		ml = new MouseAdapter() {
			public void mouseClicked(MouseEvent e) {
				if(noneSelected || selectedGraphics != null) {
					if(strokeColorChooserDialog == null) {
						strokeColorChooser = new JColorChooser(strokeColorPanel.getBackground());
						strokeColorChooserDialog = JColorChooser.createDialog(thisPanel,
							"Stroke Color", true, strokeColorChooser, finalAl2, null);
					}
					else
						strokeColorChooser.setColor(strokeColorPanel.getBackground());
					strokeColorChooserDialog.show();
				}
			}
		};
		this.strokeColorPanel.addMouseListener(ml);

		box.add(this.strokeColorPanel);
	}

	public void setNoneSelected() {
		this.filledCheckBox.setEnabled(true);
		this.strokedCheckBox.setEnabled(true);

		this.noneSelected = true;
	}

	public void setWrongTypeSelected() {
		this.filledCheckBox.setSelected(false);
		this.filledCheckBox.setEnabled(false);

		this.strokedCheckBox.setSelected(false);
		this.strokedCheckBox.setEnabled(false);

		this.noneSelected = false;
	}

	public void setRightTypeSelected() {
		ShapeGraphic graphic = (ShapeGraphic)super.selectedGraphics.elementAt(0);

		Color fill = graphic.getFillColor();
		Color stroke = graphic.getStrokeColor();

		if(fill != null) {
			this.fillColorPanel.setBackground(fill);
			this.filledCheckBox.setSelected(true);
		} else
			this.filledCheckBox.setSelected(false);

		if(stroke != null) {
			this.strokeColorPanel.setBackground(stroke);
			this.strokedCheckBox.setSelected(true);
		} else
			this.strokedCheckBox.setSelected(false);

		boolean fillValue = true;
		boolean strokeValue = true;
		for(int i = 1; i < super.selectedGraphics.size(); i++) {
			graphic = (ShapeGraphic)super.selectedGraphics.elementAt(i);

			if(graphic.getFillColor() == null && fill != null)
				fillValue = false;
			else if(graphic.getFillColor() != null && fill == null)
				fillValue = false;
			else if(graphic.getFillColor() != null && !graphic.getFillColor().equals(fill))
				fillValue = false;

			if(graphic.getStrokeColor() == null && stroke != null)
				strokeValue = false;
			else if(graphic.getStrokeColor() != null && stroke == null)
				strokeValue = false;
			else if(graphic.getStrokeColor() != null && !graphic.getStrokeColor().equals(stroke))
				strokeValue = false;
		}

		if(fillValue)
			this.filledCheckBox.setEnabled(true);
		else {
			this.filledCheckBox.setEnabled(false);
			this.filledCheckBox.setSelected(true);
		}

		if(strokeValue)
			this.strokedCheckBox.setEnabled(true);
		else {
			this.strokedCheckBox.setEnabled(false);
			this.strokedCheckBox.setSelected(true);
		}

		this.noneSelected = false;
	}

	protected void setFillColor() {
		if(super.selectedGraphics != null) {
			if (filledCheckBox.isSelected()) {
				Iterator iter = selectedGraphics.iterator();
				while (iter.hasNext()) {
					ShapeGraphic graphic = (ShapeGraphic) iter.next();
					graphic.setFillColor(fillColorPanel.getBackground());
				}
			}
			else {
				Iterator iter = selectedGraphics.iterator();
				while (iter.hasNext()) {
					ShapeGraphic graphic = (ShapeGraphic) iter.next();
					graphic.setFillColor(null);
				}
			}
		}
	}

	protected void setStrokeColor() {
		if(super.selectedGraphics != null) {
			if (strokedCheckBox.isSelected()) {
				Iterator iter = selectedGraphics.iterator();
				while (iter.hasNext()) {
					ShapeGraphic graphic = (ShapeGraphic) iter.next();
					graphic.setStrokeColor(strokeColorPanel.getBackground());
				}
			}
			else {
				Iterator iter = selectedGraphics.iterator();
				while (iter.hasNext()) {
					ShapeGraphic graphic = (ShapeGraphic) iter.next();
					graphic.setStrokeColor(null);
				}
			}
		}
	}

	public Color getFillColor() {
		if(this.filledCheckBox.isSelected())
			return this.fillColorPanel.getBackground();
		else
			return null;
	}

	public Color getStrokeColor() {
		if(this.strokedCheckBox.isSelected())
			return this.strokeColorPanel.getBackground();
		else
			return null;
	}

}