package org.convey;

import java.awt.*;

import java.awt.font.*;

import java.awt.geom.*;

import java.awt.event.*;

import org.jdom.*;

import java.util.*;

/**
 * @todo Pressing <Enter> creates a newline after placing in edit mode.
 * @author Andy Ames
 * @version 0.2
 */
public class TextGraphic extends Graphic
		implements KeyListener {

	protected Color textColor;

	/**
	 * Locaton baseline of first line of text
	 */
	protected Point origin = new Point();

	/**
	 * Text padding that provides spacing between the text and the bounds.
	 */
	protected Insets insets = new Insets(3, 3, 3, 3);

	/**
	 * Complete text buffer
	 */
	protected StringBuffer text = new StringBuffer("\n");

	protected String beforeEdit = "\n";

	/**
	 * Vector of text layouts. Each element is of type TextLayout.
	 */
	protected transient Vector lines = new Vector();

	protected int cursor = -1;

	protected int cursorLine = -1;

	protected int cursorLineIndex = -1;

	public TextGraphic() {
		this.name = "text1";
	}

	public TextGraphic(GraphicPanel graphicPanel) {
		super(graphicPanel);
		this.name = "text1";
		this.textColor = this.graphicPanel.getCanvasPanel().getConveyFrame().getTextColor();
		updateCache();
		setCursor(0);
	}

	public TextGraphic(GraphicPanel graphicPanel, Point p) {
		super(graphicPanel);
		this.name = "text1";
		this.textColor = this.graphicPanel.getCanvasPanel().getConveyFrame().getTextColor();
		origin = (Point)p.clone();
		updateCache();
		setCursor(0);
	}

	/**
	 * Paints the TextGraphic object onto the Graphics2D object.
	 */
	public void paint(Graphics2D g2d) {
		if(this.textColor == null)
			return;

		g2d.setColor(this.textColor);

		FontMetrics fm = g2d.getFontMetrics();

		int x = this.origin.x + this.insets.left;
		int y = this.origin.y + this.insets.top;

		Iterator iter = this.lines.iterator();
		while(iter.hasNext()) {
			TextLayout layout = (TextLayout)iter.next();
			layout.draw(g2d, x, y);

			y += fm.getHeight();
		}

		if(this.cursorLine != -1) {
			TextLayout layout = (TextLayout)this.lines.elementAt(this.cursorLine);

			Rectangle layoutBounds = new Rectangle(this.origin.x + this.insets.left,
																						 this.origin.y + fm.getHeight() * this.cursorLine - fm.getAscent() + this.insets.top,
																						 this.bounds.width - this.insets.left - this.insets.right,
																						 fm.getHeight());
			Shape[] carets = layout.getCaretShapes(this.cursorLineIndex, layoutBounds);
			g2d.translate(layoutBounds.x, 0);
			g2d.draw(carets[0]);
			g2d.translate(-layoutBounds.x, 0);
		}
	}

	public void paintSelection(Graphics2D g2d) {
		g2d.setColor(Color.gray);
		g2d.setStroke(new BasicStroke());
		g2d.draw(super.getBounds());
	}

	/**
	 * Sets the state of the graphic instance. The possible states are CREATING,
	 * DORMANT, SELECTED, VERTEX_MOVING, and DELETE. Note that if the state is
	 * changing to or from DORMANT, we must repaint to hide/show this graphic's
	 * selection.
	 */
	public void setState(int state) {
		super.setState(state);
		if(super.getState() != Graphic.SELECTED && this.cursor != -1)
			this.setCursor(-1);
	}

	public Element toElement() {
		Element element = super.toElement();
		element.setName("text");

		element.setAttribute("x", String.valueOf(this.origin.x));
		element.setAttribute("y", String.valueOf(this.origin.y));
		element.setAttribute("fill", this.toFillColor());

		element.addContent(this.text.toString());

		return element;
	}

	public void fromElement(Element element) {
		super.fromElement(element);

		this.origin.x = Integer.parseInt(element.getAttributeValue("x"));
		this.origin.y = Integer.parseInt(element.getAttributeValue("y"));
		this.textColor = this.stringToColor(element.getAttributeValue("fill"));

		this.text = new StringBuffer(element.getText());

		this.updateCache();
		this.repaint();
	}

	public boolean hitVertex(Point p) {
		return false;
	}

	public void fitToRect(Rectangle r) {
		// do nothing

		//super.fitToRect(r);
	}

	public void mousePressed(MouseEvent e) {
		if(this.cursor == -1)
			super.mousePressed(e);
		else
			edit(e.getPoint());
	}

	public void mouseClicked(MouseEvent e) {

	}

	public void mouseMoved(MouseEvent e) {

	}

	public void mouseDragged(MouseEvent e) {
		if(this.cursor == -1)
			super.mouseDragged(e);
	}

	public void mouseReleased(MouseEvent e) {
		if(this.cursor == -1)
			super.mouseReleased(e);
	}

	public void edit(Point p) {
		Graphics2D g2d = (Graphics2D)super.graphicPanel.getGraphics();
		FontMetrics fm = g2d.getFontMetrics();

		TextLayout line = (TextLayout)this.lines.elementAt(0);

		if(p.y < this.origin.y - fm.getAscent() + fm.getHeight() + this.insets.top) {
			TextHitInfo info = line.hitTestChar(p.x - this.origin.x - this.insets.left,
																					p.y - this.origin.y - this.insets.top);
			if(info.getInsertionIndex() == line.getCharacterCount())
				setCursor(info.getInsertionIndex() - 1);
			else
				setCursor(info.getInsertionIndex());
			return;
		}

		int index = line.getCharacterCount() + 1;

		for(int i = 1; i < this.lines.size(); i++) {
			line = (TextLayout)this.lines.elementAt(i);

			if((p.y < this.origin.y - fm.getAscent() + fm.getHeight() * i + fm.getHeight() + this.insets.top) ||
				 (i == this.lines.size() - 1)) {

				TextHitInfo info = line.hitTestChar(p.x - this.origin.x - this.insets.left,
																						p.y - this.origin.y - this.insets.top);
				if(info.getInsertionIndex() == line.getCharacterCount())
					setCursor(index + info.getInsertionIndex() - i - 1);
				else
					setCursor(index + info.getInsertionIndex() - i);
				return;
			}

			index += line.getCharacterCount() + 1;
		}
	}

	public void updateCache() {
		Graphics2D g2d = (Graphics2D)super.graphicPanel.getGraphics();
		if(g2d == null)
			return;

		FontMetrics fm = g2d.getFontMetrics();

		this.lines.clear();

		int y = this.origin.y - fm.getAscent();
		int x = this.origin.x;
		int width = 0;
		int height = 0;

		int st = 0;
		int en = 0;
		while(st < this.text.length()) {
			while(en < this.text.length()) {
				if(this.text.charAt(en) == '\n')
					break;
				en++;
			}

			StringBuffer run = new StringBuffer(this.text.substring(st, en));
			run.append(' ');

			en++;

			TextLayout layout = new TextLayout(run.toString(), g2d.getFont(), g2d.getFontRenderContext());

			lines.add(layout);

			height += fm.getHeight();

			width = (int)Math.max(width, layout.getBounds().getWidth());

			st = en;
		}

		super.bounds = new Rectangle(x, y,
																 width + this.insets.left + this.insets.right,
																 height + this.insets.top + this.insets.bottom);

		super.currDirty = (Rectangle)super.bounds.clone();
		super.currDirty.grow(5, 5);
		super.pathHotSpot = (Rectangle)super.bounds.clone();
	}

	public void translate(int x, int y) {
		super.translate(x, y);
		origin.translate(x, y);
		super.repaint();
	}

	public void setOrigin(Point p) {
		this.translate(p.x - this.origin.x, p.y - this.origin.y);
	}

	public void setText(String str) {
		if(str == null || str.length() == 0) {
			this.text.setLength(0);
			this.text.append('\n');
		} else if(str.charAt(str.length() - 1) == '\n') {
			this.text.setLength(0);
			this.text.append(str);
		} else {
			this.text.setLength(0);
			this.text.append(str);
			this.text.append("\n");
		}

		this.updateCache();
		super.repaint();
	}

	public int getCursor() {
		return this.cursor;
	}

	public void setCursor(int cursor) {
		if(cursor < 0) {
			if(super.getState() == Graphic.CREATING)
				super.setState(Graphic.SELECTED);
			else
				super.graphicPanel.addGraphicEvent(new TextGraphicEvent(this, this.beforeEdit, this.text.toString(),
						this.textColor, this.textColor));
			this.cursor = -1;
			this.cursorLine = -1;
			this.cursorLineIndex = -1;
		} else if(cursor >= this.text.length()) {
			this.cursor = this.text.length() - 1;
			this.cursorLine = this.lines.size();
			this.cursorLine = ((TextLayout)this.lines.elementAt(this.lines.size() - 1)).getCharacterCount();
		} else {
			this.cursor = cursor;

			this.cursorLine = 0;
			int lastNewline = 0;
			for(int i = 0; i < this.cursor; i++) {
				if(this.text.charAt(i) == '\n') {
					this.cursorLine++;
					lastNewline = i + 1;
				}
			}

			this.cursorLineIndex = this.cursor - lastNewline;
		}

		this.beforeEdit = this.text.toString();
		super.forceRepaint();
	}

	/**
	 * Invoked when a key has been typed. This event occurs when a key press is
	 * followed by a key release.
	 */
	public void keyTyped(KeyEvent e) {
		if(this.cursor == -1)
			return;

		char ch = e.getKeyChar();
		if(!Character.isISOControl(ch) || ch == '\n') {
			this.text.insert(this.cursor, ch);
			this.updateCache();
			this.setCursor(this.cursor + 1);
		}
	}

	/**
	 * Invoked when a key has been pressed.
	 */
	public void keyPressed(KeyEvent e) {
		if(this.cursor == -1) {
			if(e.getKeyCode() == KeyEvent.VK_ENTER)
				this.setCursor(0);
			return;
		}

		Graphics2D g2d = (Graphics2D)super.graphicPanel.getGraphics();
		FontMetrics fm = g2d.getFontMetrics();

		switch(e.getKeyCode()) {
			case KeyEvent.VK_LEFT:
				if(this.cursor == 0)
					return;

				if(e.isControlDown()) {
					boolean eatWhite = false;
					int newCursor = this.cursor;

					if(Character.isWhitespace(this.text.charAt(this.cursor - 1)))
						eatWhite = true;

					while(newCursor > 0 && eatWhite) {
						newCursor--;
						if(Character.isWhitespace(this.text.charAt(newCursor)))
							eatWhite = false;
					}

					while(newCursor > 0) {
						newCursor--;
						if(newCursor > 0 && Character.isWhitespace(this.text.charAt(newCursor - 1)))
							break;
					}

					this.setCursor(newCursor);
				} else {
					if(this.cursor > 0)
						this.setCursor(this.cursor - 1);
				}
				break;
			case KeyEvent.VK_RIGHT:
				if(this.cursor == this.text.length() - 1)
					return;

				if(e.isControlDown()) {
					boolean eatChars = false;
					int newCursor = this.cursor;

					if(!Character.isWhitespace(this.text.charAt(this.cursor)))
						eatChars = true;

					while(newCursor < this.text.length() - 1 && eatChars) {
						newCursor++;
						if(Character.isWhitespace(this.text.charAt(newCursor)))
							eatChars = false;
					}

					while(newCursor < this.text.length() - 1) {
						newCursor++;
						if(newCursor < this.text.length() - 1 && !Character.isWhitespace(this.text.charAt(newCursor)))
							break;
					}

					this.setCursor(newCursor);
				} else {
					if(this.cursor < this.text.length() - 1)
						this.setCursor(this.cursor + 1);
				}
				break;
			case KeyEvent.VK_UP:
				if(this.cursorLine == 0)
					return;

				TextLayout layout = (TextLayout)this.lines.elementAt(this.cursorLine);

				Rectangle layoutBounds = new Rectangle(this.origin.x + this.insets.left,
																							 this.origin.y + fm.getHeight() * this.cursorLine - fm.getAscent() + this.insets.top,
																							 this.bounds.width - this.insets.left - this.insets.right,
																							 fm.getHeight());
				Shape[] carets = layout.getCaretShapes(this.cursorLineIndex,
																							 layoutBounds);
				Rectangle caretBounds = carets[0].getBounds();
				this.edit(new Point(caretBounds.x + layoutBounds.x,
														caretBounds.y - fm.getHeight()));
				break;
			case KeyEvent.VK_DOWN:
				if(this.cursorLine == this.lines.size() - 1)
					return;

				layout = (TextLayout)this.lines.elementAt(this.cursorLine);

				layoutBounds = new Rectangle(this.origin.x + this.insets.left,
																		 this.origin.y + fm.getHeight() * this.cursorLine - fm.getAscent() + this.insets.top,
																		 this.bounds.width - this.insets.left - this.insets.right,
																		 fm.getHeight());
				carets = layout.getCaretShapes(this.cursorLineIndex, layoutBounds);
				caretBounds = carets[0].getBounds();
				this.edit(new Point(caretBounds.x + layoutBounds.x, caretBounds.y + fm.getHeight()));
				break;
			case KeyEvent.VK_HOME:
				if(e.isControlDown())
					this.setCursor(0);
				else
					this.setCursor(this.cursor - this.cursorLineIndex);
				break;
			case KeyEvent.VK_END:
				if(e.isControlDown())
					this.setCursor(this.text.length() - 1);
				else
					this.setCursor(this.cursor + ((TextLayout)this.lines.elementAt(this.cursorLine)).getCharacterCount() - this.cursorLineIndex - 1);
				break;
			case KeyEvent.VK_ESCAPE:
				this.setCursor(-1);
				break;
			case KeyEvent.VK_BACK_SPACE:
				if(this.cursor == 0)
					return;
				this.text.deleteCharAt(this.cursor - 1);
				this.updateCache();
				this.setCursor(this.cursor - 1);
				break;
			case KeyEvent.VK_DELETE:
				if(this.cursor == this.text.length() - 1)
					return;
				this.text.deleteCharAt(this.cursor);
				this.updateCache();
				this.forceRepaint();
				break;
		}
	}

	/**
	 * Invoked when a key has been released.
	 */
	public void keyReleased(KeyEvent e) {

	}

	public void setTextColor(Color color) {
		if(color == null)
			return;

		Color origColor = this.textColor;

		this.textColor = color;
		super.forceRepaint();

		if(!super.graphicPanel.isIncoming() && super.getState() != Graphic.CREATING)
			super.graphicPanel.addGraphicEvent(new TextGraphicEvent(this, this.text.toString(), this.text.toString(),
					origColor, this.textColor));
	}

	public Color getTextColor() {
		return this.textColor;
	}

	public String toFillColor() {
		return colorToString(this.textColor);
	}

}